import sys
from notario.exceptions import Invalid, SchemaError
from notario.utils import (is_callable, sift, is_empty, re_sort, is_not_empty,
                           data_item, safe_repr, ensure)
from notario.normal import Data, Schema
from notario.validators import cherry_pick


class Validator(object):

    def __init__(self, data, schema, defined_keys=None):
        if defined_keys:
            schema = cherry_pick(schema)
        self.data = Data(data, schema).normalized()
        self.schema = Schema(data, schema).normalized()

    def validate(self):
        if self.data == {} and self.schema:
            msg = 'has no data to validate against schema'
            reason = 'an empty dictionary object was provided'
            raise Invalid(None, {}, msg=msg, reason=reason, pair='value')
        self.traverser(self.data, self.schema, [])

    def traverser(self, data, schema, tree):
        """
        Traverses the dictionary, recursing onto itself if
        it sees appropriate key/value pairs that indicate that
        there is a need for more validation in a branch below us.
        """
        if hasattr(schema, '__validator_leaf__'):
            return schema(data, tree)

        if hasattr(schema, 'must_validate'):  # cherry picking?
            if not len(schema.must_validate):
                reason = "must_validate attribute must not be empty"
                raise SchemaError(data, tree, reason=reason)
            data = sift(data, schema.must_validate)

        schema = self.sanitize_optionals(data, schema, tree)
        self.is_alpha_ordered(data, schema, tree)

        validated_indexes = []
        skip_missing_indexes = getattr(schema, 'must_validate', False)

        if len(data) < len(schema):
            # we have missing required items in data, but we don't know
            # which ones so find what may fail:
            data_keys = [v[1] for v in data.values()]
            schema_keys = [v[1] for v in schema.values()]

            def enforce_once(data_keys, schema_key):
                # XXX Go through all the data keys and try and see if they pass
                # validation against the schema. At this point it is impossible
                # to know which data key corresponds to what schema key
                # (because schema keys can be a function/callable) so it is
                # a *very* naive way to try and detect which one might be
                # missing
                for data_key in data_keys:
                    failed = None
                    try:
                        enforce(data_key, schema_key, tree, pair='key')
                        return
                    except Invalid:
                        failed = data_key, schema_key

                    if failed:
                        return failed

            for schema_key in schema_keys:
                failure = enforce_once(data_keys, schema_key)
                if failure:
                    _, failed_schema_key = failure
                    msg = "required key in data is missing: %s" % str(failed_schema_key)
                    raise Invalid(None, tree, reason=msg, pair='key')

        for index in range(len(data)):
            self.length_equality(data, schema, index, tree)
            key, value = data[index]
            skey, svalue = schema[index]
            tree.append(key)

            # Validate the key before anything, to prevent recursing
            self.key_leaf(data[index], schema[index], tree)

            # If a dict is a value we need to recurse.
            # XXX Should we check isinstance(value, ndict) ?
            if isinstance(value, dict) and len(value):
                self.traverser(value, svalue, tree)
            else:
                self.value_leaf(data[index], schema[index], tree)
            if tree:
                tree.pop()

            validated_indexes.append(index)

        # XXX There is a chance we might have missing items from
        # the incoming data that are labeled as required from the schema
        # we should make sure *here* that we account for that and raise
        # the appropriate exception. Since the loop finished and everything
        # seems to have passed, this lack of check will give false positives.
        missing_indexes = set(schema.keys()).difference(validated_indexes)
        if missing_indexes:
            if skip_missing_indexes:
                return
            for i in missing_indexes:
                if not hasattr(schema[i], 'is_optional'):
                    required_key = schema[i][0]
                    tree.append('item[%s]' % i)
                    msg = "required item in schema is missing: %s" % str(required_key)
                    raise Invalid(required_key, tree, reason=msg, pair='key')


    def key_leaf(self, data, schema, tree):
        """
        The deepest validation we can make in any given circumstance for a key.
        Does not recurse, it will just receive both values and the tree,
        passing them on to the :fun:`enforce` function.
        """
        key, value = data
        schema_key, schema_value = schema
        enforce(key, schema_key, tree, 'key')

    def value_leaf(self, data, schema, tree):
        """
        The deepest validation we can make in any given circumstance for
        a value. Does not recurse, it will just receive both values and the
        tree, passing them on to the :fun:`enforce` function.
        """
        key, value = data
        schema_key, schema_value = schema

        if hasattr(schema_value, '__validator_leaf__'):
            return schema_value(value, tree)
        enforce(value, schema_value, tree, 'value')

    def is_alpha_ordered(self, data, normalized_schema, tree):
        keys = []
        indexes = normalized_schema.keys()
        for index in indexes:
            key = normalized_schema[index][0]
            if isinstance(key, str):
                keys.append(key)
            elif hasattr(key, '_object') :
                if isinstance(key._object, str):
                    keys.append(key._object)

        sorted_keys = sorted(keys)
        if keys != sorted_keys:
            for index, key in enumerate(keys):
                if key != sorted_keys[index]:
                    raise SchemaError(
                        keys, [key],
                        reason='schema item is not alphabetically ordered'
                    )

    def length_equality(self, data, schema, index, tree):
        try:
            data = data[index]
            try:
                schema = schema[index]
            except KeyError:
                if not hasattr(schema, 'must_validate'):
                    reason = 'has unexpected item in data: %s' % data_item(data)
                    raise Invalid(None, tree, msg=reason, reason=reason, pair='value')
        except (KeyError, TypeError):
            if not hasattr(schema, 'must_validate'):
                reason = "has less items in schema than in data"
                raise SchemaError(data, tree, reason=reason)
        if hasattr(schema, '__validator_leaf__'):
            return

        if len(data) != len(schema):
            raise SchemaError(data, tree, reason='length did not match schema')

    def sanitize_optionals(self, data, schema, tree):
        schema_key_map = {}
        try:
            for number, value in schema.items():
                schema_key_map[number] = getattr(value[0], '_object', value[0])
        except AttributeError:  # maybe not a dict?
            self.length_equality(data, schema, 0, tree)

        optional_keys = {}
        for k, v in schema.items():
            try:
                key = getattr(v[0], '_object')
                if key:
                    optional_keys[k] = key
            except AttributeError:
                pass

        data_keys = [v[0] for k, v in data.items()]

        for number, value in optional_keys.items():
            if value not in data_keys:
                del schema[number]
        if not schema and is_not_empty(data):
            msg = "unexpected extra items"
            raise Invalid(schema, tree, reason=msg)
        return re_sort(schema)


class BaseItemValidator(object):

    def __init__(self, data, schema, tree=None, index=None, name=None):
        self.data = data
        self.schema = schema
        self.tree = tree or []
        self.index = index or 0
        self.name = name

    def validate(self):
        self.traverser(self.data, self.schema, self.tree)

    def traverser(self, data, schema, tree):
        if len(data) < self.index:
            reason = "has not enough items to select from"
            raise SchemaError(data, tree, reason=reason)
        self.leaves(data, schema, tree)


class IterableValidator(BaseItemValidator):
    """
    The iterable validator allows the definition of a single schema that can be
    run against any number of items in a given data structure
    """

    def data_sanity(self, data, tree=None):
        if not isinstance(data, list):
            name = self.name or 'IterableValidator'
            reason = 'expected a list but got %s' % safe_repr(data)
            msg = 'did not pass validation against callable: %s' % name
            raise Invalid('', tree or [], msg=msg, reason=reason, pair='value')

    def leaf(self, index):
        self.data_sanity(self.data, tree=self.tree)
        self.enforce(self.data, self.schema, index, self.tree)

    def leaves(self, data, schema, tree):
        self.data_sanity(data, tree=tree)
        for item_index in range(self.index, len(data)):
            self.enforce(data, schema, item_index, tree)

    def enforce(self, data, schema, item_index, tree):
        # yo dawg, a recursive validator within a recursive validator anyone?
        if is_callable(schema) and hasattr(schema, '__validator_leaf__'):
            return schema(data[item_index], tree)
        if isinstance(data[item_index], dict) and isinstance(schema, tuple):
            try:
                _validator = Validator(data[item_index], schema)
                _validator.validate()
            except Invalid:
                e = sys.exc_info()[1]
                tree.append('list[%s]' % item_index)
                tree.extend(e.path)
                raise Invalid(e.schema_item, tree, reason=e._reason, pair='value')

            # FIXME this is utterly redundant, and also happens in
            # RecursiveValidator
            except SchemaError:
                e = sys.exc_info()[1]
                tree.extend(e.path)
                raise SchemaError('', tree, reason=e._reason, pair='value')

        elif isinstance(schema, tuple) and not isinstance(data[item_index], (tuple, dict)):
            raise SchemaError(data, tree, reason='iterable contains single items, schema does not')
        else:
            try:
                if is_callable(schema):
                    schema(data[item_index])
                else:
                    ensure(data[item_index] == schema)
            except AssertionError:
                reason = sys.exc_info()[1]
                tree.append('list[%s]' % item_index)
                raise Invalid(schema, tree, reason=reason, pair='item')


class RecursiveValidator(BaseItemValidator):
    """
    The recursive validator allows the definition of a single schema that can
    be run against any number of items in a given data structure
    """

    def leaf(self, index):
        self.enforce(self.data, self.schema, index, self.tree)

    def leaves(self, data, schema, tree):
        for item_index in range(self.index, len(data)):
            self.enforce(data, schema, item_index, tree)

    def enforce(self, data, schema, item_index, tree):
        # yo dawg, a recursive validator within a recursive validator anyone?
        if is_callable(schema) and hasattr(schema, '__validator_leaf__'):
            return schema(data, tree)
        try:
            _validate = Validator({}, self.schema)
            _validate.data = {0: data[item_index]}
            _validate.validate()
        except Invalid:
            e = sys.exc_info()[1]
            tree.extend(e.path)
            raise Invalid(e.schema_item, tree, pair='value', msg=e._msg, reason=e._reason)
        except SchemaError:
            e = sys.exc_info()[1]
            tree.extend(e.path)
            raise SchemaError('', tree, reason=e._reason, pair='value')


def enforce(data_item, schema_item, tree, pair):
    schema_is_optional = hasattr(schema_item, 'is_optional')
    if is_callable(schema_item) and not schema_is_optional:
        try:
            schema_item(data_item)
        except AssertionError:
            e = sys.exc_info()[1]
            if pair == 'value':
                tree.append(data_item)
            raise Invalid(schema_item, tree, reason=e, pair=pair)
    else:
        try:
            if schema_is_optional:
                if is_empty(data_item):  # we received nothing here
                    return
                ensure(data_item == schema_item())
            else:
                ensure(data_item == schema_item)
        except AssertionError:
            e = sys.exc_info()[1]
            if pair == 'value':
                tree.append(data_item)
            raise Invalid(schema_item, tree, reason=e, pair=pair)


def validate(data, schema, defined_keys=False):
    """
    Main entry point for the validation engine.

    :param data: The incoming data, as a dictionary object.
    :param schema: The schema from which data will be validated against
    """
    if isinstance(data, dict):
        validator = Validator(data, schema, defined_keys=defined_keys)
        validator.validate()
    else:
        raise TypeError('expected data to be of type dict, but got: %s' % type(data))





class cherry_pick(tuple):
    """
    This is not really a validator, but will allow one to make
    sure just certain keys (as defined in ``must_validate`` are
    validated. The engine will discard all others that are not
    defined here.

    It should be used in the same manner as when constructing
    a schema, but instead of a regular tuple, this should be used.

    .. note::
        If ``must_validate`` is not defined with any items, it will
        generate them with the items on the tuple passed in if possible, unless
        the tuple created is not a key value pair.

    """
    must_validate = ()

    def __init__(self, _tuple):
        try:
            self.must_validate = tuple([key for key, value in _tuple])
        except ValueError:  # single k/v pair
            if len(_tuple) == 2:
                self.must_validate = (_tuple[0],)
        return super(cherry_pick, self).__init__()


class Hybrid(object):
    """
    Validators in Notario are usually meant to validate very specific, granular
    *values* or *items*, but there is no easy way of trying to validate
    a single value or to enforce a schema if for some reason the validator
    doesn't know if it will get a single value or an object.

    For example, if the constraint was something like: "value can be either an
    integer or an array with integers", it would not be possible to juggle that
    with other validators.

    The Hybrid validator helps with that as it is able to accept both
    a validator and a schema. If it receives something that looks like a single
    value (like a boolean, integer or string) it will use the validator,
    otherwise, it will trigger Notario's engine along with the schema provided
    at initialization.

    This is how the previous scenario would be represented::

    .. doctest:: Hybrid

        >>> from notario.validators import Hybrid
        >>> from notario import validate, ensure
        >>> def validator(value):
            ... ensure(isinstance(value, bool))
            ...
        >>> schema = ('a', Hybrid(validator, ('a', 1)))
        >>> data = {'a', False}
        >>> validate(data, schema)
        >>> # if data changes, the validator accomodates for it
        >>> data = {'a': {'a': 1}}
        >>> vaidate(data, schema)

    .. warning::
        Only if the argument passed in to this validator is of type ``list`` or
        ``dict`` will the schema be enforced, otherwise it will fall back to
        using the validator.

    """

    __validator_leaf__ = True

    def __init__(self, validator, schema):
        self.validator = validator
        self.schema = schema

    def __call__(self, value, *args):
        if isinstance(value, (dict, list)):
            from notario.validators.recursive import RecursiveValidator
            validator = RecursiveValidator(value, self.schema, *args)
            validator.validate()
        else:
            try:
                tree = args[0]
            except IndexError:
                tree = []
            from notario.engine import enforce
            enforce(value, self.validator, tree, pair='value')



from notario.exceptions import Invalid
from notario.utils import safe_repr, expand_schema, is_schema
from notario.engine import RecursiveValidator


class BasicRecursiveValidator(object):
    """
    Base class for recursive validators, can be sub-classed
    for other type of recursive validators but should not be
    used directly.
    """

    __validator_leaf__ = True

    def __init__(self, schema):
        self.schema = schema


class AnyObject(BasicRecursiveValidator):
    """
    Go over all the items in an dict object and make sure that at least
    one of the items validates correctly against the schema provided.
    If no items pass it raises ``Invalid``.


    .. testsetup:: anyobject

        from notario import validate
        from notario.validators.recursive import AnyObject

    Example usage for single values::

        data = {'foo': {{'a':10}, {'b':20}, {'c':40}}}
        schema = ('foo', AnyObject(('a', 10)))
        validate(data, schema)

    Example usage for other data structures::

        data = {'foo': [{'b': 10}, {'a': 1}]}
        schema = ('foo', AnyObject(('a', 1))
        validate(data, schema)

    When a single item in the array fails to pass against the validator's
    schema it stops further iteration and it will raise an error like:


    .. doctest:: anyobject

        >>> data = {'foo': {'a':{'a':10}, 'b':{'a':20}, 'c':{'a':20}}}
        >>> schema = ('foo', AnyObject(('a', ('a', 90))))
        >>> validate(data, schema)
        Traceback (most recent call last):
        ...
        Invalid: foo -> a  did not contain any valid objects against callable: AnyObject

    In this particular validator, it remembers on what key of the dict object
    the failure was created and it goes even further giving the key and value
    of the object it went against.
    """

    def __call__(self, data, tree):
        schema = expand_schema(self.schema)
        index = len(data) - 1
        validator = RecursiveValidator(data, schema, [], index=index)
        for item_index in range(len(data)):
            try:
                return validator.leaf(item_index)
            except Invalid:
                if tree:
                    tree.pop
                pass

        msg = "did not contain any valid objects against callable: %s" % self.__class__.__name__
        raise Invalid(schema, tree, pair='value', msg=msg)


class AllObjects(BasicRecursiveValidator):
    """
    For all the objects contained in a dictionary apply the schema passed in
    to the validator.
    If a single item object fails, it raises ``Invalid``.


    .. testsetup:: allobjects

        from notario import validate
        from notario.validators.recursive import AllObjects

    Example usage for single values::

        data = {'foo': {{'a':10}, {'a':20}, {'a':20}}}
        schema = ('foo', AllObjects(('a', 20)))
        validate(data, schema)

    Example usage for other data structures::

        data = {'foo': [{'a': 1}, {'a': 1}]}
        schema = ('foo', AllObjects(('a', 1))
        validate(data, schema)

    When a single item in the array fails to pass against the validator's
    schema it stops further iteration and it will raise an error like:


    .. doctest:: allobjects

        >>> data = {'foo': {'a':{'a':10}, 'b':{'a':20}, 'c':{'a':20}}}
        >>> schema = ('foo', AllObjects(('a', ('a', 90))))
        >>> validate(data, schema)
        Traceback (most recent call last):
        ...
        Invalid: -> foo -> a -> a -> 10  did not match 90

    In this particular validator, it remembers on what key of the dict object
    the failure was created and it goes even further giving the key and value
    of the object it went against.


    """

    def __call__(self, data, tree):
        schema = expand_schema(self.schema)
        validator = RecursiveValidator(data, schema, tree)
        validator.validate()


class MultiRecursive(object):

    """
    .. testsetup:: *

        from notario import validate
        from notario.validators.recursive import MultiRecursive

    This validator is useful when there is a need for validating any number of
    schemas for a given data set.

    If the requirement is, for example, to validate either ``string: string``
    or ``string: int`` this will not be possible with any of the other
    validators in Notario because they all assume a rule that must dominate
    everything. Otherwise, it is required to specify the expectation.

    In the case that the object to validate complies with the above
    requirement, this validator can accept **any number** of schemas as
    arguments, so if one schema fails, the next one will be tried until all the
    schemas are applied for a given item. The ``MultiRecursive`` validator will
    look like this in order to pass the incoming data:

    .. doctest:: multirecursive

        >>> data = {'main': {'foo': 'bar'}}
        >>> schema = ('main', MultiRecursive(('foo', 1), ('foo', 'bar')))
        >>> validate(data, schema)

    Because we can't be sure what the data may hold we are forced to define
    different rules and apply them so that they can pass. If we repeat the
    expectation but an invalid value comes along, the validator will naturally
    fail:

    .. doctest:: mutlirecursive_fail

        >>> data = {'main': {'foo': False}}
        >>> schema = ('main', MultiRecursive(('foo', 1), ('foo', 'bar')))
        >>> validate(data, schema)
        Traceback (most recent call last):
        ...
        Invalid: -> foo -> False did not match 'bar'

    """
    __validator_leaf__ = True

    def __init__(self, *schemas):
        for schema in schemas:
            if not is_schema(schema):
                raise TypeError("got a non schema argument: %s" % safe_repr(schema))
        self.schemas = schemas

    def __call__(self, data, tree):
        """
        Go through each item in data against the first schema in the arguments.
        If that fails try each subsequent schema until it passes. Even if
        schemas are failing to apply to that given item, consume all the
        available ones until at least one passes.

        If nothing passes, the last error is raised.

        :param data: Incoming data from the validator engine in normalized dict
        form.
        :param tree: The traversing tree up to this point, always passed in.
        :raises Invalid: If none of the schemas can validate the data.
        """
        first_schema = expand_schema(self.schemas[0])
        index = len(data) - 1
        validator = RecursiveValidator(data, first_schema, [], index=index)
        for item_index in range(len(data)):
            try:
                validator.leaf(item_index)
            except Invalid:
                self.itemized_validation(validator, item_index)

    def itemized_validation(self, validator, item_index):
        error = None

        for schema in self.schemas:
            try:
                validator.schema = expand_schema(schema)
                validator.tree = []
                return validator.leaf(item_index)
            except Invalid as e:
                error = e

        if error is not None:
            raise error



import warnings


def is_callable(data):
    if hasattr(data, '__call__'):
        return True
    return False


def safe_repr(obj):
    """
    Try to get ``__name__`` first, ``__class__.__name__`` second
    and finally, if we can't get anything acceptable, fallback
    to user a ``repr()`` call.
    """
    name = getattr(obj, '__name__', getattr(obj.__class__, '__name__'))
    if name == 'ndict':
        name = 'dict'
    return name or repr(obj)


# Backwards compatibility
def optional(validator):
    from notario import decorators
    msg = "import optional from notario.decorators, not from utils"
    warnings.warn(msg, DeprecationWarning, stacklevel=2)
    return decorators.optional(validator)


class ndict(dict):
    """
    Used by Notario so that it can slap attributes to the object
    when it is created, something that regular dictionaries will not
    allow to do.
    """
    pass


def re_sort(data):
    """
    A data with keys that are not enumerated sequentially will be
    re sorted and sequentially ordered.

    For example::

        >>> data = {16: ('1', 'b'), 3: ('1', 'a')}
        >>> re_sort(data)
        >>> {0: ('1', 'a'), 1: ('1', 'b')}
    """
    keys = sorted(data.keys())
    new_data = {}
    for number, key in enumerate(keys):
        new_data[number] = data[key]
    return new_data


def sift(data, required_items=None):
    """
    Receive a ``data`` object that will be in the form
    of a normalized structure (e.g. ``{0: {'a': 0}}``) and
    filter out keys that match the ``required_items``.
    """
    required_items = required_items or []
    new_data = {}
    for k, v in data.items():
        if v[0] in required_items:
            new_data[k] = v
            continue
        for required_item in required_items:
            key = getattr(required_item, '_object', False)
            if key:
                if v[0] == key:
                    new_data[k] = v

    return re_sort(new_data)


def is_empty(value):
    try:
        return len(value) == 0
    except TypeError:
        return False


def is_not_empty(value):
    return not is_empty(value)


def is_nested_tuple(value):
    if len(value) == 2 and isinstance(value[1], tuple): # nested tuple
        return True
    return False


def data_item(data):
    """
    When trying to return a meaningful error about an unexpected data item
    we cannot just `repr(data)` as that could show a gigantic data struture.

    This utility should try to get the key of the first item or the single item
    in the data structure.
    """
    if isinstance(data, ndict):
        # OK, we have something that looks like {0: ('a', 'b')}
        # or something that is a regular dictionary
        # so try to return 'a' regardless of the length
        for item in data:
            return repr(data[item][0])
    elif isinstance(data, dict):
        for item in data:
            return repr(data[item])
    elif isinstance(data, list):
        return repr(data[0])
    return repr(data)


def forced_leaf_validator(func):
    """
    Some decorators may need to use this if doing anything related
    with a dictionary value. Since callables are usually validating
    single values (e.g. a string or a boolean) and not a dictionary
    per-se because Notario will already normalize the data.
    """
    func.__validator_leaf__ = True
    return func


def expand_schema(schema):
    if hasattr(schema, '__delayed__'):
        return schema()
    return schema


def is_schema(_object):
    """
    A helper to determine if we are indeed dealing with what it seems to be
    a schema.
    """
    if hasattr(_object, '__delayed__') or isinstance(_object, tuple):
        return True
    return False


def ensure(assertion, message=None):
    """
    Checks an assertion argument for truth-ness. Will return ``True`` or
    explicitly raise ``AssertionError``. This is to deal with environments
    using ``python -O` or ``PYTHONOPTIMIZE=``.

    :param assertion: some value to evaluate for truth-ness
    :param message: optional message used for raising AssertionError
    """
    message = message or assertion

    if not assertion:
        raise AssertionError(message)

    return True



"""
A basic module to provide both Python2 and Python3 support
on the same source
"""
import sys

if sys.version_info >= (3, 0): # pragma: no cover
    basestring = str
else:
    basestring = basestring



from pytest import raises
from notario.validators import types
from notario.validators import recursive
from notario.exceptions import Invalid
from notario.normal import Data, Schema


class TestAllObjects(object):

    def test_all_objects_fail(self):
        data = {0:('a', '1'), 1:('b', 2), 2:('c', '3')}
        schema = ((types.string, types.integer))
        with raises(Invalid) as exc:
            any_object = recursive.AllObjects(schema)
            any_object(data, [])
        msg = '-> a -> 1 did not pass validation against callable: integer'
        error = exc.value.args[0]
        assert msg in error
        assert 'not of type int' in error

    def test_all_objects_pass(self):
        data = {0:('a', 1), 1:('b', 2), 2:('c', 3)}
        schema = ((types.string, types.integer))
        any_object = recursive.AllObjects(schema)
        assert any_object(data, []) is None


class TestAnyObject(object):

    def test_any_object_fail_all(self):
        data = {0:('a', '1'), 1:('b', '2'), 2:('c', '3')}
        schema = ((types.string, types.integer))
        with raises(Invalid) as exc:
            any_object = recursive.AnyObject(schema)
            any_object(data, [])
        msg = '-> top level did not contain any valid objects against callable: AnyObject'
        assert exc.value.args[0] == msg

    def test_any_object_pass_first(self):
        data = {0:('a', '1'), 1:('b', 2), 2:('c', 3)}
        schema = ((types.string, types.string))
        any_object = recursive.AnyObject(schema)
        assert any_object(data, []) is None

    def test_any_object_pass_second(self):
        data = {0:('a', 1), 1:('b', '2'), 2:('c', 3)}
        schema = ((types.string, types.string))
        any_object = recursive.AnyObject(schema)
        assert any_object(data, []) is None

    def test_any_object_pass_last(self):
        data = {0:('a', 1), 1:('b', 2), 2:('c', '3')}
        schema = ((types.string, types.string))
        any_object = recursive.AnyObject(schema)
        assert any_object(data, []) is None


class TestMultiSchema(object):

    def test_pass_single_data_item(self):
        data = Data({'a': 2}, {}).normalized()
        schemas = (('a', 2), ('b', 1))

        multi = recursive.MultiRecursive(*schemas)
        assert multi(data, []) is None

    def test_fail_on_non_callable(self):
        with raises(TypeError):
            recursive.MultiRecursive(False)

    def test_pass_two_data_items(self):
        data = Data({'a': 2, 'b': 1}, {}).normalized()
        schemas = (('a', 2), ('b', 1))

        multi = recursive.MultiRecursive(*schemas)
        assert multi(data, []) is None

    def test_pass_on_second_schema(self):
        data = Data({'b': 1}, {}).normalized()
        multi = recursive.MultiRecursive(('a', 2), ('b', 1))
        assert multi(data, []) is None

    def test_fail_on_last_schema(self):
        data = Data({'a': 2, 'b': 1}, {}).normalized()
        schemas = (('a', 2), ('b', 2))

        multi = recursive.MultiRecursive(*schemas)
        with raises(Invalid) as exc:
            multi(data, [])
        assert '1 did not match 2' in exc.value.args[0]

    def test_pass_on_third_schema(self):
        data = Data({'a': 2, 'b': 1}, {}).normalized()
        schemas = (('a', 2), ('b', 2), ('b', 1))

        multi = recursive.MultiRecursive(*schemas)
        assert multi(data, []) is None

    def test_pass_single_on_third_schema(self):
        data = Data({'z': 2}, {}).normalized()
        schemas = (('a', 2), ('b', 2), ('z', 2))

        multi = recursive.MultiRecursive(*schemas)
        assert multi(data, []) is None

    def test_fail_single_on_third_schema(self):
        data = Data({'z': 2}, {}).normalized()
        schemas = (('a', 2), ('b', 2), ('z', 1))

        multi = recursive.MultiRecursive(*schemas)
        with raises(Invalid) as exc:
            multi(data, [])
        assert 'z -> 2 did not match 1' in exc.value.args[0]




from pytest import raises
from notario import engine
from notario.exceptions import Invalid, SchemaError
from notario.validators import recursive, iterables, types
from notario.decorators import optional
from notario.tests import util


class TestEnforce(object):

    def test_callable_failure_raises_invalid(self):
        def bad_callable(v): assert False
        with raises(Invalid) as exc:
            engine.enforce('1', bad_callable, ['1'], 'key')
        error = exc.value.args[0]
        assert '-> 1 key did not pass validation against callable: bad_callable' in error

    def test_callable_passes_with_flying_colors(self):
        def cool_callable(v): pass
        result = engine.enforce('1', cool_callable, ['1'], 'key')
        assert result is None

    def test_equal_params_pass_with_flying_colors(self):
        result = engine.enforce('1', '1', ['1'], 'key')
        assert result is None

    def test_unequal_params_raise_invalid(self):
        with raises(Invalid) as exc:
            engine.enforce(1, 2, ['1'], 'key')
        error = exc.value.args[0]
        assert '-> 1 key did not match 2' in error

    def test_callable_with_messages_are_passed_on(self):
        def callable_message(v): assert False, "this is completely False"
        with raises(Invalid) as exc:
            engine.enforce(1, callable_message, ['1'], 'key')
        result = util.assert_message(exc.value.reason)
        assert result  == "this is completely False"


class TestValidator(object):

    def test_validate_top_level_keys(self):
        data = {'a': 1, 'b': 2}
        schema = (('a', 1), ('c', 2))
        with raises(Invalid) as exc:
            validator = engine.Validator(data, schema)
            validator.validate()

        error = exc.value.args[0]
        assert  "-> b key did not match 'c'" in error

    def test_validate_length_equality(self):
        data = {'a': 1, 'b': 2}
        schema = (('a', 1, 2), ('c', 2))
        with raises(SchemaError) as exc:
            validator = engine.Validator(data, schema)
            validator.validate()

        error = exc.value.args[0]
        assert  '-> top level length did not match schema' in error

    def test_validate_length_equality_less_items(self):
        data = {'a': 1, 'b': 2, 'c':'c'}
        schema = (('a', 1), ('b', 2))
        with raises(Invalid) as exc:
            validator = engine.Validator(data, schema)
            validator.validate()

        error = exc.value.args[0]
        assert "top level has unexpected item in data: ('c', 'c')" in error

    def test_validate_length_equality_returns(self):
        data = {0:('a', 1)}
        class _Schema(object):
            __validator_leaf__ = True
        schema = {0:_Schema()}
        validator = engine.Validator({}, {})
        result = validator.length_equality(data, schema, 0, [])
        assert result is None

    def test_validate_top_level_values(self):
        data = {'a': 1, 'b': 2}
        schema = (('a', 1), ('b', 1))
        with raises(Invalid) as exc:
            validator = engine.Validator(data, schema)
            validator.validate()

        error = exc.value.args[0]
        assert  '-> b -> 2 did not match 1' in error

    def test_validate_value_nested_dictionaries(self):
        data = {'a': 1, 'b': {'a': 2, 'b' : 1}}
        schema = (('a', 1), ('b', (('a', 2), ('b', 2))))
        with raises(Invalid) as exc:
            validator = engine.Validator(data, schema)
            validator.validate()

        error = exc.value.args[0]
        assert  '-> b -> b -> 1 did not match 2' in error

    def test_validate_key_doubly_nested_dictionaries(self):
        data = {'a': 1, 'b': {'a': 2, 'b' : {'a':'a'}}}
        schema = (('a', 1), ('b', (('a', 2), ('b', ('a', 'b')))))
        with raises(Invalid) as exc:
            validator = engine.Validator(data, schema)
            validator.validate()

        error = exc.value.args[0]
        assert  "-> b -> b -> a -> a did not match 'b'" in error

    def test_validate_key_nested_dictionaries(self):
        data = {'a': 1, 'b': {'a': 2, 'b' : 1}}
        schema = (('a', 1), ('b', (('b', 2), ('b', 1))))
        with raises(Invalid) as exc:
            validator = engine.Validator(data, schema)
            validator.validate()

        error = exc.value.args[0]
        assert  "-> b -> a key did not match 'b'" in error

    def test_validate_value_nested_arrays(self):
        data = {'a': 1, 'b': {'a': 2, 'b' : [1, 2, 3]}}
        schema = (('a', 1), ('b', (('a', 2), ('b', [1, 1, 3]))))
        with raises(Invalid) as exc:
            validator = engine.Validator(data, schema)
            validator.validate()

        error = exc.value.args[0]
        assert  '-> b -> b -> [1, 2, 3] did not match [1, 1, 3]' in error

    def test_validate_key_nested_arrays(self):
        data = {'a': 1, 'b': {'a': 2, 'b' : [1, 2, 3]}}
        schema = (('a', 1), ('b', (('a', 2), ('b', [1, 1, 3]))))
        with raises(Invalid) as exc:
            validator = engine.Validator(data, schema)
            validator.validate()

        error = exc.value.args[0]
        assert  '-> b -> b -> [1, 2, 3] did not match [1, 1, 3]' in error

    def test_validate_multiple_items_as_values(self):
        data = {'a': 1, 'b': {'a': 2, 'b' : 1, 'd':1, 'c':2}}
        schema = (('a', 1), ('b', (('a', 2), ('b', 1), ('c', 2), ('d', 2))))
        with raises(Invalid) as exc:
            validator = engine.Validator(data, schema)
            validator.validate()

        error = exc.value.args[0]
        assert  '-> b -> d -> 1 did not match 2' in error in error


class TestValidatorLeaves(object):
    """
    The Validator object would behave differently with
    validator classes that have  the __validator_leaf__
    """

    def test_traverser_returns_the_iterable_leaf_if_seen(self):
        data = {'a': {'b': [1, 1, 1, 1]}}
        schema = ('a', ('b', iterables.AllItems(2)))
        with raises(Invalid) as exc:
            validator = engine.Validator(data, schema)
            validator.validate()

        error = exc.value.args[0]
        assert  '-> a -> b -> list[0] item did not match 2' in error

    def test_traverser_returns_the_recursive_leaf_if_seen(self):
        data = {'a': {'b': [1, 1, 1, 1]}}
        schema = ('a', recursive.AllObjects(('b', [1, 1, 1, 2])))
        with raises(Invalid) as exc:
            validator = engine.Validator(data, schema)
            validator.validate()

        error = exc.value.args[0]
        assert  '-> a -> b -> [1, 1, 1, 1] did not match [1, 1, 1, 2]' in error


class TestRecursiveValidator(object):

    def test_bad_index_number(self):
        data = {'a': {'a':'a', 'b':'b'}}
        schema = ('a', 'a')
        recursive_validator = recursive.RecursiveValidator(data, schema, index=100)
        with raises(SchemaError) as exc:
            recursive_validator.validate()

        error = exc.value.args[0]
        assert  '-> top level has not enough items to select from' in error

    def test_deal_with_recursion(self):
        data = {'a': {'a': 'a', 'b': {'a': 'b', 'c': 'c', 'd': 1}}}
        schema = ('a', (('a', 'a'), ('b', recursive.AllObjects((types.string, types.string)))))
        with raises(Invalid) as exc:
            validator = engine.Validator(data, schema)
            validator.validate()

        error = exc.value.args[0]
        assert '1 did not pass validation against callable: string' in error
        assert 'not of type string' in error


class TestIterableValidator(object):

    def test_not_enough_items_in_data(self):
        data = {0: ('a', 'b')}
        schema = {0: ('a', 'b')}
        iter_validator = engine.IterableValidator(data, schema, index=100)
        with raises(SchemaError) as exc:
            iter_validator.validate()
        error = exc.value.args[0]
        assert '-> top level has not enough items to select from' in error

    def test_validate_nested_array_first(self):
        data = [{'a':'a'}, {'b': 'b'}]
        schema = (types.string, 'b') # note how this is not normalized
        iter_validator = engine.IterableValidator(data, schema, index=0)
        with raises(Invalid) as exc:
            iter_validator.validate()

        error = exc.value.args[0]
        assert  "-> list[0] -> a -> a did not match 'b'" in error

    def test_validate_nested_array(self):
        data = [{'a':'b'}, {'b': 'c'}]
        schema = (types.string, 'b') # note how this is not normalized
        iter_validator = engine.IterableValidator(data, schema, index=0)
        with raises(Invalid) as exc:
            iter_validator.validate()

        error = exc.value.args[0]
        assert  "-> list[1] -> b -> c did not match 'b'" in error

    def test_report_schema_errors(self):
        data = [1,2,3,4]
        schema = (types.string, ('b', 'a'))
        iter_validator = engine.IterableValidator(data, schema, index=0)
        with raises(SchemaError) as exc:
            iter_validator.validate()
        error = exc.value.args[0]
        assert "iterable contains single items, schema does not" in error

    def test_invalid_on_non_list(self):
        data = {'foo': [{'a':'b'}, {'b': 'c'}]}
        schema = (types.string, 'b') # note how this is not normalized
        iter_validator = engine.IterableValidator(data, schema, index=0)
        with raises(Invalid) as exc:
            iter_validator.validate()
        error = exc.value.args[0]
        assert "top level did not pass validation against callable: IterableValidator" in error


class TestOptionalKeys(object):

    def test_optional_key_passes(self):
        data = {'a': 1, 'c': 3}
        schema = (('a', 1), (optional('b'), 2), ('c', 3))
        validator = engine.Validator(data, schema)
        assert validator.validate() is None

    def test_optional_key_raises(self):
        data = {'a': 1, 'b': 3, 'c': 3}
        schema = (('a', 1), (optional('b'), 2), ('c', 3))
        validator = engine.Validator(data, schema)
        with raises(Invalid) as exc:
            validator.validate()
        exc_msg = str(exc.value)
        assert '3 did not match 2' in exc_msg
        assert 'b -> 3' in exc_msg


class TestValidate(object):

    def test_refuses_non_dicts(self):
        with raises(TypeError):
            engine.validate(['a list'], ('a', 'b'))



from notario._compat import basestring
from notario.utils import is_callable


class NotarioException(Exception):

    def __init__(self, schema_item, path, reason=None, pair='key', msg=None):
        self.schema_item = schema_item
        self.path = path
        self._reason = reason
        self._pair = pair
        self._msg = msg
        Exception.__init__(self, self.__str__())

    def __str__(self):
        return self._get_message()

    def _format_path(self, use_pair=True):
        message = ""
        for key in self.path:
            if isinstance(key, basestring):
                if key == '':
                    key = "'%s'" % key
            accessed_key = '-> %s ' % key
            message += accessed_key
        if use_pair and self._pair != 'value':
            return message + self._pair
        return message or "-> top level"

    def _formatted_reason(self):
        if self.reason:
            return " (%s)" % self.reason
        return ''

    def _format_message(self):
        reason = self._formatted_reason()
        if self._msg:
            return self._msg
        if self.schema_item is None:
            return "did not match schema %s" % reason
        if is_callable(self.schema_item):
            msg = "did not pass validation against callable: %s"\
                  "%s" % (self.schema_item.__name__, reason)
        else:
            msg = "did not match %s%s" % (repr(self.schema_item), reason)
        return msg

    @property
    def reason(self):
        try:
            return self._reason.args[0]
        except IndexError:
            return self._reason
        except AttributeError:
            return self._reason or ''


class Invalid(NotarioException):
    """
    This Exception class is used only by the :class:`Validator`
    class to provide a clear message when a validation mismatch occurs
    while traversing the configuration tree.
    """

    def _get_message(self):
        path = self._format_path()
        message = self._format_message()
        if path.endswith(' '):
            return "%s%s" % (path, message)
        return "%s %s" % (path, message)


class NestedInvalid(Invalid):
    """
    When a validator is deeply nested, (e.g. a recursive inside an iterable)
    catching exceptions for better reporting is very difficult because of how
    most are caught and re-written to accomodate a tree.  This exception will
    probably be useful only to those types of validators, that need to raise an
    ``Invalid`` exception that should not be re-written in any way.
    """
    pass


class SchemaError(NotarioException):

    def _get_message(self):
        msg = "%s %s" % (self._format_path(use_pair=False), self._reason)
        return msg


class Skip(Exception):
    """
    This Exception class is used for ``optional`` decorators that fail
    to get the right key for a given data.
    """
    pass



"""
Iterable validators for array objects only. They provide a way of
applying a schema to any given items in an array.
"""
from notario.exceptions import Invalid, SchemaError
from notario.engine import IterableValidator
from notario.utils import is_callable, safe_repr, expand_schema, is_schema


class BasicIterableValidator(object):
    """
    Base class for iterable validators, can be sub-classed
    for other type of iterable validators but should not be
    used directly.
    """

    __validator_leaf__ = True

    def __init__(self, schema):
        self.schema = schema

    def safe_type(self, data, tree):
        """
        Make sure that the incoming data complies with the class type we
        are expecting it to be. In this case, classes that inherit from this
        base class expect data to be of type ``list``.
        """
        if not isinstance(data, list):
            name = self.__class__.__name__
            msg = "did not pass validation against callable: %s" % name
            reason = 'expected a list but got %s' % safe_repr(data)
            raise Invalid(self.schema, tree, reason=reason, pair='value', msg=msg)


class AnyItem(BasicIterableValidator):
    """
    Go over all the items in an array and make sure that at least
    one of the items validates correctly against the schema provided.
    If no items pass it raises ``Invalid``.

    .. note::
        It only works on arrays, otherwise it will raise a ``SchemaError``

    .. testsetup::

        from notario import validate
        from notario.validators.iterables import AnyItem

    Example usage for single values::

        data = {'foo' : [10, 30, 50]}
        schema = ('foo', AnyItem(50))
        validate(data, schema)

    Example usage for other data structures::

        data = {'foo': [{'a': 1}, {'b': 2}]}
        schema = ('foo', AnyItem(('b', 2))
        validate(data, schema)

    When a single item in the array matches correctly against the validator's
    schema it stops further iteration and the validation passes. Otherwise it
    will raise an error like:


    .. doctest::

        >>> data = {'foo': [{'a': 1}, {'b': 2}]}
        >>> schema = ('foo', AnyItem(('c', 4)))
        >>> validate(data, schema)
        Traceback (most recent call last):
        ...
        Invalid: -> foo -> list[]  did not contain any valid items matching ('c', 4)

    """

    def __call__(self, data, tree):
        schema = expand_schema(self.schema)
        self.safe_type(data, tree)
        index = len(data) - 1
        validator = IterableValidator(data, schema, [], index=index, name='AnyItem')
        for item_index in range(len(data)):
            try:
                return validator.leaf(item_index)
            except Invalid:
                pass

        tree.append('list[]')
        if is_callable(schema):
            msg = "did not contain any valid items against callable: %s" % schema.__name__
        else:
            msg = "did not contain any valid items matching %s" % repr(schema)
        raise Invalid(schema, tree, pair='value', msg=msg)


class AllItems(BasicIterableValidator):
    """
    For all the items in an array apply the schema passed in to the validator.
    If a single item fails, it raises ``Invalid``.

    .. note::
        It only works on arrays, otherwise it will raise a ``SchemaError``

    .. testsetup:: allitems

        from notario import validate
        from notario.validators.iterables import AllItems

    Example usage for single values::

        data = {'foo' : [10, 10, 10]}
        schema = ('foo', AllItems(10))
        validate(data, schema)

    Example usage for other data structures::

        data = {'foo': [{'a': 1}, {'a': 1}]}
        schema = ('foo', AllItems(('a', 1))
        validate(data, schema)

    When a single item in the array fails to pass against the validator's
    schema it stops further iteration and it will raise an error like:


    .. doctest:: allitems

        >>> data = {'foo': [{'a': 1}, {'a': 2}]}
        >>> schema = ('foo', AllItems(('a', 1)))
        >>> validate(data, schema)
        Traceback (most recent call last):
        ...
        Invalid: -> foo -> list[1] -> a -> 2  did not match 1

    In this particular validator, it remembers on what index of the array the
    failure was created and it goes even further giving the key and value of
    the object it went against.

    """

    def __call__(self, data, tree):
        schema = expand_schema(self.schema)
        self.safe_type(data, tree)
        validator = IterableValidator(data, schema, tree, name='AllItems')
        validator.validate()


class MultiIterable(object):

    """
    .. testsetup:: *

        from notario import validate
        from notario.validators.iterables import MultiIterable

    This validator is useful when there is a need for validating any number of
    schemas for a given data set.

    If the requirement is, for example, to validate either ``string: string``
    or ``string: int`` this will not be possible with any of the other
    validators in Notario because they all assume a rule that must dominate
    everything. Otherwise, it is required to specify the expectation.

    In the case that the object to validate complies with the above
    requirement, this validator can accept **any number** of schemas as
    arguments, so if one schema fails, the next one will be tried until all the
    schemas are applied for a given item. The ``MultiRecursive`` validator will
    look like this in order to pass the incoming data:

    .. doctest:: multiiterable

        >>> data = {'main': {'foo': 'bar'}}
        >>> schema = ('main', MultiIterable(('foo', 1), ('foo', 'bar')))
        >>> validate(data, schema)

    Because we can't be sure what the data may hold we are forced to define
    different rules and apply them so that they can pass. If we repeat the
    expectation but an invalid value comes along, the validator will naturally
    fail:

    .. doctest:: mutlirecursive_fail

        >>> data = {'main': [{'foo': False}]}
        >>> schema = ('main', MultiIterable(('foo', 1), ('foo', 'bar')))
        >>> validate(data, schema)
        Traceback (most recent call last):
        ...
        Invalid: -> foo -> False did not match 'bar'

    """
    __validator_leaf__ = True

    def __init__(self, *schemas):
        for schema in schemas:
            if not is_schema(schema):
                raise TypeError("got a non schema argument: %s" % safe_repr(schema))
        self.schemas = schemas

    def __call__(self, data, tree):
        """
        Go through each item in data against the first schema in the arguments.
        If that fails try each subsequent schema until it passes. Even if
        schemas are failing to apply to that given item, consume all the
        available ones until at least one passes.

        If nothing passes, the last error is raised.

        :param data: Incoming data from the validator engine in normalized dict
        form.
        :param tree: The traversing tree up to this point, always passed in.
        :raises Invalid: If none of the schemas can validate the data.
        """
        first_schema = expand_schema(self.schemas[0])
        index = len(data) - 1
        validator = IterableValidator(data, first_schema, tree, index=index, name='MultiIterable')

        for item_index in range(len(data)):
            try:
                validator.leaf(item_index)
            except (SchemaError, Invalid):
                self.itemized_validation(validator, item_index)

    def itemized_validation(self, validator, item_index):
        error = None

        for schema in self.schemas:
            try:
                validator.schema = expand_schema(schema)
                validator.tree = []
                return validator.leaf(item_index)
            except (SchemaError, Invalid) as e:
                error = e

        if error is not None:
            raise error



"""
Basic type validators
"""
from functools import wraps
from notario._compat import basestring
from notario.exceptions import Invalid
from notario.utils import is_callable, forced_leaf_validator, ensure


def string(_object):
    """
    Validates a given input is of type string.

    Example usage::

        data = {'a' : 21}
        schema = (string, 21)

    You can also use this as a decorator, as a way to check for the
    input before it even hits a validator you may be writing.

    .. note::
        If the argument is a callable, the decorating behavior will be
        triggered, otherwise it will act as a normal function.
    """
    if is_callable(_object):
        _validator = _object

        @wraps(_validator)
        def decorated(value):
            ensure(isinstance(value, basestring), "not of type string")
            return _validator(value)
        return decorated
    ensure(isinstance(_object, basestring), "not of type string")


def boolean(_object):
    """
    Validates a given input is of type boolean.

    Example usage::

        data = {'a' : True}
        schema = ('a', boolean)

    You can also use this as a decorator, as a way to check for the
    input before it even hits a validator you may be writing.

    .. note::
        If the argument is a callable, the decorating behavior will be
        triggered, otherwise it will act as a normal function.

    """
    if is_callable(_object):
        _validator = _object

        @wraps(_validator)
        def decorated(value):
            ensure(isinstance(value, bool), "not of type boolean")
            return _validator(value)
        return decorated
    ensure(isinstance(_object, bool), "not of type boolean")


@forced_leaf_validator
def dictionary(_object, *args):
    """
    Validates a given input is of type dictionary.

    Example usage::

        data = {'a' : {'b': 1}}
        schema = ('a', dictionary)

    You can also use this as a decorator, as a way to check for the
    input before it even hits a validator you may be writing.

    .. note::
        If the argument is a callable, the decorating behavior will be
        triggered, otherwise it will act as a normal function.

    """
    error_msg = 'not of type dictionary'
    if is_callable(_object):
        _validator = _object

        @wraps(_validator)
        def decorated(value):
            ensure(isinstance(value, dict), error_msg)
            return _validator(value)
        return decorated
    try:
        ensure(isinstance(_object, dict), error_msg)
    except AssertionError:
        if args:
            msg = 'did not pass validation against callable: dictionary'
            raise Invalid('', msg=msg, reason=error_msg, *args)
        raise


def array(_object):
    """
    Validates a given input is of type list.

    Example usage::

        data = {'a' : [1,2]}
        schema = ('a', array)

    You can also use this as a decorator, as a way to check for the
    input before it even hits a validator you may be writing.

    .. note::
        If the argument is a callable, the decorating behavior will be
        triggered, otherwise it will act as a normal function.

    """
    if is_callable(_object):
        _validator = _object

        @wraps(_validator)
        def decorated(value):
            ensure(isinstance(value, list), "not of type array")
            return _validator(value)
        return decorated
    ensure(isinstance(_object, list), "not of type array")


def integer(_object):
    """
    Validates a given input is of type int..

    Example usage::

        data = {'a' : 21}
        schema = ('a', integer)

    You can also use this as a decorator, as a way to check for the
    input before it even hits a validator you may be writing.

    .. note::
        If the argument is a callable, the decorating behavior will be
        triggered, otherwise it will act as a normal function.
    """
    if is_callable(_object):
        _validator = _object

        @wraps(_validator)
        def decorated(value):
            ensure(isinstance(value, int), "not of type int")
            return _validator(value)
        return decorated
    ensure(isinstance(_object, int), "not of type int")



import re


class Linker(list):
    """
    This list-like object will receive key/value pairs as regexes with
    accompanying failure error messages and upon a call, it will apply them
    sequentially so that a user can understand at what point in the regex the
    failure happened with a given value.

    .. note::
        Direct use of this class is discouraged as the functionality is exposed
        through the :def:`chain` helper function.
    """

    def __init__(self, regexes, prepend_negation=None):
        if prepend_negation is None:
            self.prepend_negation = False
        else:
            self.prepend_negation = prepend_negation
        self._sanity(regexes)
        self.prepend_negation = prepend_negation
        self.regexes = regexes
        self._build()

    def _sanity(self, items):
        try:
            for item in items:
                pattern, reason = item
        except ValueError:
            raise TypeError('arguments must be key value pairs')

    def _build(self):
        self.append(re.compile(self.regexes[0][0]))
        for number in range(1, len(self.regexes)):
            new_regex = self._get_regex(number)
            self.append(re.compile(new_regex))

    def _get_regex(self, number):
        items = []
        for item, _ in self.regexes[:number + 1]:
            items.append(item)
        return ''.join(items)

    def _get_reason(self, item_number):
        reason = self.regexes[item_number][1]
        if self.prepend_negation:
            return "does not %s" % self.regexes[item_number][1]
        return reason

    def __call__(self, value):
        for count, regex in enumerate(self):
            if not regex.match(value):
                raise AssertionError(self._get_reason(count))


def chain(*regexes, **kwargs):
    """
    A helper function to interact with the regular expression engine
    that compiles and applies partial matches to a string.

    It expects key value tuples as arguments (any number of them) where the
    first pair is the regex to compile and the latter is the message to display
    when the regular expression does not match.

    The engine constructs partial regular expressions from the input and
    applies them sequentially to find the exact point of failure and allowing
    the ability to return a meaningful message.

    Because adding negation statements like "does not..." can become
    repetitive, the function defaults to ``True`` to include the option to
    prepend the negative.

    For example, this is what would happen with a failing regex::

        >>> rx = chain((r'^\d+', 'start with a digit'))
        >>> rx('foo')
        Traceback (most recent call last):
        ...
        AssertionError: does not start with a digit

    If there is no need for prepending the negation, the keyword argument will
    need to set it as ``False``::

        >>> rx = chain((r'^\d+', 'it should start with a digit'),
        ...            prepend_negation=False)
        >>> rx('foo')
        Traceback (most recent call last):
        ...
        AssertionError: it should start with a digit
    """

    prepend_negation = kwargs.get('prepend_negation', True)
    return Linker(regexes, prepend_negation=prepend_negation)



from pytest import raises

from notario.normal import Data
from notario import utils
from notario.decorators import delay


class TestIsCallable(object):

    def setup(self):
        def fake_callable(): pass
        self.fake_callable = fake_callable

    def test_is_callable(self):
        result = utils.is_callable(self.fake_callable)
        assert result is True

    def test_is_not_callable(self):
        result = utils.is_callable(1)
        assert result is False


class TestOptionalDecorator(object):

    def fake_validator(self, value):
        return True

    def test_has_no_value(self):
        result = utils.optional(self.fake_validator)('')
        assert result is None

    def test_value(self):
        result = utils.optional(self.fake_validator)('some value!')
        assert result is True

    def test_decorator_is_marked_as_optional(self):
        result = utils.optional('')
        assert result.is_optional is True

    def test_value_has_string(self):
        result = utils.optional('value')
        assert result() == 'value'


class RequiredItem(object):
    _object = ''


class TestSift(object):

    def test_no_required_items(self):
        data = Data({'a': 1, 'b': 2}, {}).normalized()
        result = utils.sift(data)
        assert result == {}

    def test_get_one_item_back(self):

        data = Data({'a': 1, 'b': 2}, {}).normalized()
        result = utils.sift(data, ['b'])
        assert result == {0: ('b', 2)}

    def test_required_item_is_function_with_object(self):
        item = RequiredItem()
        item._object = 'a'
        data = Data({'a': 1, 'b': 2}, {}).normalized()
        result = utils.sift(data, [item])
        assert result == {0: ('a', 1)}


class TestReSort(object):

    def test_keys_will_reset_to_zero(self):
        data = {16: ('a', 1), 4: ('b', 1), 3: ('a',1)}
        result = utils.re_sort(data)
        assert result == {0: ('a', 1), 1: ('b', 1), 2: ('a', 1)}


class TestIsEmpty(object):

    def test_not_any_valid_structure_is_false(self):
        result = utils.is_empty(False)
        assert result is False

    def test_valid_structures_are_false_when_empty(self):
        for structure in [[], {}, '']:
            assert utils.is_empty(structure) is True


class TestIsNestedTuple(object):

    def test_is_nested_tuple(self):
        value = (1, (1, 2))
        assert utils.is_nested_tuple(value) is True

    def test_is_single_tuple(self):
        value = (1,)
        assert utils.is_nested_tuple(value) is False

    def test_second_value_is_not_typle(self):
        value = ('a', 'b')
        assert utils.is_nested_tuple(value) is False


class TestDataItem(object):

    def setup(self):
        self.ndict = utils.ndict()

    def test_return_key_from_ndict(self):
        self.ndict[0] = ('a', 'b')
        assert utils.data_item(self.ndict) == "'a'"

    def test_return_key_from_non_ndict(self):
        data = {'b': 'b'}
        assert utils.data_item(data) == "'b'"

    def test_return_first_item_in_list(self):
        data = ['a', 'b', 'c']
        assert utils.data_item(data) == "'a'"

    def test_fallback_to_repr_of_obj(self):
        assert utils.data_item(self.ndict) == "{}"


class TestExpandSchema(object):

    def test_not_delayed(self):
        schema = ('a', 'b')
        assert utils.expand_schema(schema) == schema

    def test_is_delayed(self):
        @delay
        def my_schema():
            return 'a', 'b'
        assert utils.expand_schema(my_schema) == ('a', 'b')


class TestIsSchema(object):

    def test_if_is_tuple(self):
        assert utils.is_schema(('a', 'b')) is True

    def test_is_delayed(self):
        schema = delay(lambda x: True)
        assert utils.is_schema(schema) is True

    def test_is_not_tuple(self):
        assert utils.is_schema([]) is False


class TestEnsure(object):

    def test_ensure(self):
        assert utils.ensure(1 == 1)

    def test_ensure_raises_assertionerror(self):
        with raises(AssertionError):
            utils.ensure(0 == 1)



from pytest import raises
from notario.validators import Hybrid
from notario.exceptions import Invalid
from notario.decorators import optional
from notario import validate


def validator(x):
    assert x, 'fail'


class TestHybrid(object):

    def test_use_validator_passes(self):
        schema = ()
        hybrid = Hybrid(validator, schema)
        assert hybrid(1) is None

    def test_use_validator_fails(self):
        schema = ()
        hybrid = Hybrid(validator, schema)
        with raises(Invalid) as exc:
            hybrid(0)
        error = exc.value.args[0]
        assert '0 did not pass validation against callable' in error

    def test_use_schema_passes(self):
        schema = ('a', 1)
        hybrid = Hybrid(validator, schema)
        hybrid({0: ('a', 1)})

    def test_use_schema_fails(self):
        schema = ('a', 2)
        hybrid = Hybrid(validator, schema)
        with raises(Invalid) as exc:
            hybrid({0: ('a', 1)})
        error = exc.value.args[0]
        assert 'a -> 1 did not match 2' in error


class TestFunctional(object):

    def test_passes_single_value(self):
        sschema = (1, 2)
        schema = ('a', Hybrid(validator, sschema))
        data = {'a': 2}
        assert validate(data, schema) is None

    def test_passes_object(self):
        sschema = (1, 2)
        schema = ('a', Hybrid(validator, sschema))
        data = {'a': {1: 2}}
        assert validate(data, schema) is None

    def test_fail_object(self):
        sschema = (1, 1)
        schema = ('a', Hybrid(validator, sschema))
        data = {'a': {1: 2}}
        with raises(Invalid) as exc:
            validate(data, schema)
        error = exc.value.args[0]
        assert '1 -> 2 did not match 1' in error
        assert error.startswith('-> a -> 1')

    def test_extra_unexpected_items(self):
        optional_schema = (optional(1), 1)
        schema = ('a', Hybrid(validator, optional_schema))
        data = {'a': {'foo': 'bar'}}
        with raises(Invalid) as exc:
            validate(data, schema)
        error = exc.value.args[0]
        assert '-> a did not match {}' in error
        assert exc.value.reason == 'unexpected extra items'



from notario._compat import basestring


def assert_message(exception):
    """
    Newever py.test versions modify the AssertionError exceptions so that they
    can add more information to it when using the ``==`` operator for
    comparison.
    This helper splits on the newlines and returns the first part, providing
    a backwards compatible solution that only extracts the portion of the
    assertion that notario would normally use.

    A statement like ``assert 3 == 1, 'too big'`` would look like this in newer py.test versions::

        AssertionError(u'too big\nassert 3 == 1',)

    But before they would look like::


        AssertionError(u'too big',)
    """
    if isinstance(exception, basestring):
        arguments = exception
    else:
        arguments = exception.value.args[0]
    return arguments.split('\n')[0]



from notario import validators


class TestCherryPick(object):

    def test_single_pair_tuple(self):
        result = validators.cherry_pick(('a', 'b'))
        assert result.must_validate == ('a',)

    def test_single_items_tuples(self):
        result = validators.cherry_pick(('a', 'b', 'c', 'd'))
        assert result.must_validate == ()

    def test_empty_tuple(self):
        result = validators.cherry_pick(())
        assert result.must_validate == ()



from pytest import raises
from notario.validators import types
from notario.tests import util

#
# Most of these are just excercising the code really
# as they are just making sure isinstance returns
# what we expect, effectively testing python
#


class TestTypes(object):

    def test_string_pass(self):
        assert types.string('a string') is None

    def test_boolean_pass(self):
        assert types.boolean(False) is None

    def test_dictionary_pass(self):
        assert types.dictionary({}) is None

    def test_array_pass(self):
        assert types.array([]) is None

    def test_integer_pass(self):
        assert types.integer(1) is None

    def test_string_fail(self):
        with raises(AssertionError) as exc:
            types.string(1)
        assert exc.value.args[0] == 'not of type string'

    def test_boolean_fail(self):
        with raises(AssertionError) as exc:
            types.boolean('a string')
        assert exc.value.args[0] == 'not of type boolean'

    def test_dictionary_fail(self):
        with raises(AssertionError) as exc:
            types.dictionary([])
        assert exc.value.args[0] == 'not of type dictionary'

    def test_array_fail(self):
        with raises(AssertionError) as exc:
            types.array({})
        assert exc.value.args[0] == 'not of type array'

    def test_integer_fail(self):
        with raises(AssertionError) as exc:
            types.integer('a string')
        assert exc.value.args[0] == 'not of type int'


class TestTypesAsDecorators(object):

    def test_string_pass(self):
        @types.string
        def validate(value):
            assert value == "a string"

        assert validate('a string') is None

    def test_boolean_pass(self):
        @types.boolean
        def validate(value):
            assert value is False

        assert validate(False) is None

    def test_dictionary_pass(self):
        @types.dictionary
        def validate(value):
            assert value == {}

        assert validate({}) is None

    def test_array_pass(self):
        @types.array
        def validate(value):
            assert value == []

        assert validate([]) is None

    def test_integer_pass(self):
        @types.integer
        def validate(value):
            assert value == 1

        assert validate(1) is None

    def test_string_fail(self):
        @types.string
        def validate(value):
            assert value == 'a string' # pragma: no cover

        with raises(AssertionError) as exc:
            validate(1)
        assert exc.value.args[0] == 'not of type string'

    def test_boolean_fail(self):
        @types.boolean
        def validate(value):
            assert value is False # pragma: no cover

        with raises(AssertionError) as exc:
            validate('a string')
        assert exc.value.args[0] == 'not of type boolean'

    def test_dictionary_fail(self):
        @types.dictionary
        def validate(value):
            assert value == {} # pragma: no cover

        with raises(AssertionError) as exc:
            validate([])
        assert exc.value.args[0] == 'not of type dictionary'

    def test_array_fail(self):
        @types.array
        def validate(value):
            assert value is False # pragma: no cover

        with raises(AssertionError) as exc:
            validate({})
        assert exc.value.args[0] == 'not of type array'

    def test_integer_fail(self):
        @types.integer
        def validate(value):
            assert value == 1 # pragma: no cover

        with raises(AssertionError) as exc:
            validate('a string')
        assert exc.value.args[0] == 'not of type int'


class TestTypesDelegating(object):

    def test_string_decorated(self):
        @types.string
        def validate(value):
            assert len(value) == 1, 'too long'

        with raises(AssertionError) as exc:
            validate('123')
        result = util.assert_message(exc)
        assert result == 'too long'

    def test_boolean_decorated(self):
        @types.boolean
        def validate(value):
            assert value is False, 'not false'

        with raises(AssertionError) as exc:
            validate(True)
        result = util.assert_message(exc)
        assert result == 'not false'

    def test_dictionary_decorated(self):
        @types.dictionary
        def validate(value):
            assert len(value) == 1, 'too long'

        with raises(AssertionError) as exc:
            validate({'a': 1, 'b': 2})
        result = util.assert_message(exc)
        assert result == 'too long'

    def test_array_decorated(self):
        @types.array
        def validate(value):
            assert len(value) == 1, 'too long'

        with raises(AssertionError) as exc:
            validate([1, 2])
        result = util.assert_message(exc)
        assert result == 'too long'

    def test_integer_fail(self):
        @types.integer
        def validate(value):
            assert value == 1, 'too big'

        with raises(AssertionError) as exc:
            validate(3)

        result = util.assert_message(exc)
        assert result == 'too big'



import copy
from notario.utils import ndict, is_nested_tuple


class BaseNormalize(object):

    def __init__(self, data, schema):
        self.raw_data = copy.deepcopy(data)
        self.raw_schema = schema

    def ordered_dict(self, data_structure, use_n_dict=False):
        if use_n_dict:
            ordered = ndict()
        else:
            ordered = {}
        for number, value in enumerate(data_structure):
            ordered[number] = value
        return ordered


class Data(BaseNormalize):

    def _normalize(self, data_structure, sort=True):
        if sort:
            for k, v in data_structure.items():
                if isinstance(v, dict):
                    data_structure[k] = self._normalize(v)
            data_structure = sorted(data_structure.items())
        return self.ordered_dict(data_structure, use_n_dict=True)

    def normalized(self):
        """
        Normalizes the ``raw_data`` (a plain dictionary) and returns an ordered
        dictionary, with integers as keys for tuples.
        """
        normalized_data = self._normalize(self.raw_data)
        return normalized_data


class Schema(BaseNormalize):

    def _normalize(self, data):
        if len(data) == 2 and isinstance(data[1], tuple) or len(data) > 2:
            if not isinstance(data[0], tuple):
                new_struct = ndict({0: (data[0], self._normalize(data[1]))})
            else:
                new_struct = self.ordered_dict(data, use_n_dict=True)
            for i in range(len(new_struct)):
                value = new_struct.get(i)
                if is_nested_tuple(value):
                    new_struct[i] = (value[0], self._normalize(value[1]))
            if hasattr(data, 'must_validate'):
                new_struct.must_validate = data.must_validate
            return new_struct
        else:
            new_struct = ndict({0: data})
            if hasattr(data, 'must_validate'):
                new_struct.must_validate = data.must_validate
            return new_struct

    def normalized(self):
        normalized_data = self._normalize(self.raw_schema)
        return normalized_data



from py.test import raises
from notario import decorators


class TestInstanceOf(object):

    def test_not_of_any_valid_types(self):
        @decorators.instance_of()
        def validator(value):
            assert True  # pragma: no cover

        with raises(AssertionError) as exc:
            validator(0)
        exc_msg = str(exc.value)

        assert 'not of any valid types' in exc_msg
        assert 'list' in exc_msg
        assert 'dict' in exc_msg
        assert 'str' in exc_msg

    def test_custom_valid_types(self):
        @decorators.instance_of((str,))
        def validator(value):
            assert True # pragma: no cover

        with raises(AssertionError) as exc:
            validator({})
        exc_msg = str(exc.value)

        assert 'not of any valid types' in exc_msg
        assert 'str' in exc_msg

    def test_validate_dict_or_list(self):
        @decorators.instance_of((list,dict))
        def validator(value):
            assert True # pragma: no cover

        assert validator([]) is None
        assert validator({}) is None


    def test_custom_class(self):
        class Foo(object):
            pass

        @decorators.instance_of((Foo,))
        def validator(value):
            assert True # pragma: no cover

        with raises(AssertionError) as exc:
            validator({})
        exc_msg = str(exc.value)

        assert 'not of any valid types' in exc_msg
        assert 'Foo' in exc_msg


class TestNotEmpty(object):

    def test_not_of_any_valid_types(self):
        with raises(AssertionError) as exc:
            decorators.not_empty(False)
        errors = exc.value.args[0]
        assert 'not of any valid types' in errors

    def test_not_empty_dict(self):
        @decorators.not_empty
        def validator(value):
            assert len(value) # pragma: no cover

        with raises(AssertionError) as exc:
            validator({})
        exc_msg = str(exc.value)

        assert 'is empty' in exc_msg

    def test_not_empty_str(self):
        @decorators.not_empty
        def validator(value):
            assert len(value) # pragma: no cover

        with raises(AssertionError) as exc:
            validator("")
        exc_msg = str(exc.value)

        assert 'is empty' in exc_msg

    def test_not_empty_list(self):
        @decorators.not_empty
        def validator(value):
            assert len(value) # pragma: no cover

        with raises(AssertionError) as exc:
            validator([])
        exc_msg = str(exc.value)

        assert 'is empty' in exc_msg

    def test_non_decorator_empty_string(self):
        with raises(AssertionError) as exc:
            decorators.not_empty("")
        exc_msg = str(exc.value)

        assert 'is empty' in exc_msg

    def test_is_empty_delegates_to_validator(self):
        @decorators.not_empty
        def validator(value):
            assert len(value) == 2, 'not two'

        with raises(AssertionError) as exc:
            validator([1])
        exc_msg = str(exc.value)

        assert 'not two' in exc_msg



from pytest import raises
from notario.validators import iterables, types
from notario.exceptions import Invalid, SchemaError


class TestAnyItem(object):

    def test_any_item_fail(self):
        data = [1, 2, 6, 4, 5]
        schema = 7
        with raises(Invalid) as exc:
            any_item = iterables.AnyItem(schema)
            any_item(data, [])
        msg = '-> list[] did not contain any valid items matching 7'
        assert exc.value.args[0] == msg

    def test_tuple_schemas_cannot_run_on_single_values(self):
        data = [1, 2, 6, 4, 5]
        schema = ('a', 'b')
        with raises(SchemaError) as exc:
            any_item = iterables.AnyItem(schema)
            any_item(data, [])
        msg = "-> top level iterable contains single items, schema does not"
        assert exc.value.args[0] == msg

    def test_single_schemas_can_run_on_tuple_items(self):
        data = [{'a':'a'}, {'b':'b'}]
        schema = 'a'
        with raises(Invalid) as exc:
            any_item = iterables.AnyItem(schema)
            any_item(data, [])
        msg = "-> list[] did not contain any valid items matching 'a'"
        assert exc.value.args[0] == msg

    def test_any_item_pass(self):
        data = [1, 2, 6, 4, 5]
        schema = 6
        any_item = iterables.AnyItem(schema)
        assert  any_item(data, []) is None

    def test_any_item_pass_last_item(self):
        data = [1, 2, 3, 4, 6]
        schema = 6
        any_item = iterables.AnyItem(schema)
        assert  any_item(data, []) is None

    def test_any_item_fail_on_non_lists(self):
        data = 4
        schema = 1
        all_items = iterables.AnyItem(schema)
        with raises(Invalid) as exc:
            all_items(data, [])
        error = exc.value.args[0]
        assert  'top level did not pass validation against callable: AnyItem' in error

    def test_expecting_list_error(self):
        data = 4
        schema = 1
        all_items = iterables.AnyItem(schema)
        with raises(Invalid) as exc:
            all_items(data, [])
        assert 'expected a list but got int' == exc.value.reason



class TestAllItems(object):

    def test_all_items_pass(self):
        data = [1, 1, 1, 1]
        schema = 1
        all_items = iterables.AllItems(schema)
        assert all_items(data, []) is None

    def test_all_items_fail_on_non_lists(self):
        data = 4
        schema = 1
        all_items = iterables.AllItems(schema)
        with raises(Invalid) as exc:
            all_items(data, [])
        error = exc.value.args[0]
        assert  'top level did not pass validation against callable: AllItems' in error

    def test_expecting_list_error(self):
        data = 4
        schema = 1
        all_items = iterables.AllItems(schema)
        with raises(Invalid) as exc:
            all_items(data, [])
        assert 'expected a list but got int' == exc.value.reason

    def test_all_items_fail(self):
        data = [1, 1, 3, 1]
        schema = 1
        all_items = iterables.AllItems(schema)
        with raises(Invalid) as exc:
            all_items(data, [])
        error = exc.value.args[0]
        assert  '-> list[2] item did not match 1' in error

    def test_all_items_fail_last_item(self):
        data = [1, 1, 1, 3]
        schema = 1
        all_items = iterables.AllItems(schema)
        with raises(Invalid) as exc:
            all_items(data, [])

        error = exc.value.args[0]
        assert  '-> list[3] item did not match 1' in error

    def test_all_items_fail_uses_first_invalid(self):
        data = [3, 5, 2, 3]
        schema = 1
        all_items = iterables.AllItems(schema)
        with raises(Invalid) as exc:
            all_items(data, [])

        error = exc.value.args[0]
        assert  '-> list[0] item did not match 1' in error

    def test_fails_when_missing_items(self):
        data = [{"host": "example.com"}]
        schema = (("host", types.string),("interface", types.string))
        all_items = iterables.AllItems(schema)
        with raises(Invalid) as exc:
            all_items(data, [])
        error = exc.value.args[0]
        assert  "-> list[0] -> item[1] did not match 'interface'" in  error
        assert "(required item in schema is missing: interface)" in error



from pytest import raises
from notario import validate, ensure
from notario.exceptions import Invalid, SchemaError
from notario.validators import (iterables, recursive, types,
                                chainable, cherry_pick, Hybrid)
from notario.decorators import optional, instance_of


class TestValidate(object):

    def test_simple_validation_against_empty_data(self):
        data = {}
        schema = (('a', 'b'))

        with raises(Invalid) as exc:
            validate(data, schema)

        error = exc.value.args[0]
        assert 'top level has no data to validate against schema' in error

    def test_single_level(self):
        data = {'members': ['1']}
        schema = ('members', types.array)
        validate(data,schema)


    def test_instance_of_dict_or_list(self):
        #from notario.utils import forced_leaf_validator
        @instance_of((list, dict))
        def valid_custom(value):
            ensure(len(value) > 0)

        data_dict = {'foo' : {"some string": 1}}
        data_list = {'foo' : ['a', 'b']}
        schema = ('foo', valid_custom)
        validate(data_dict, schema)
        validate(data_list, schema)

    def test_fail_instance_of_dict(self):
        #from notario.utils import forced_leaf_validator
        @instance_of((dict,))
        def valid_custom(value):
            ensure(len(value) > 5, "%s does not have len of 5" % value)

        data_dict = {'foo' : {"some string": 1}}
        schema = ('foo', valid_custom)
        with raises(AssertionError) as exc:
            validate(data_dict, schema)
        error = exc.value.args[0]
        assert "{'some string': 1}" in error
        assert "does not have len of 5" in error

    def test_most_simple_validation(self):
        data = {'a': 'a'}
        schema = (('a', 'b'))

        with raises(Invalid) as exc:
            validate(data, schema)

        error = exc.value.args[0]
        assert  "-> a -> a did not match 'b'" in error

    def test_required_key_is_missing(self):
        data = {'a': 'a', 'c': 'c'}
        schema = (('a', 'a'), ('b', 'b'), ('c', 'c'))

        with raises(Invalid) as exc:
            validate(data, schema)

        error = exc.value.args[0]
        assert  "key did not match schema" in error
        assert  "required key in data is missing: b" in error
    def test_key_is_empty_string(self):
        data = {'a': ''}
        schema = (('a', 'b'))

        with raises(Invalid) as exc:
            validate(data, schema)

        error = exc.value.args[0]
        assert  "-> a -> '' did not match 'b'" in error

    def test_multi_pair_non_nested_last(self):
        data = {'a': 'a', 'b':'b', 'c':'c', 'd':'d'}
        schema = (('a', 'a'), ('b', 'b'), ('c', 'c'), ('d', 'a'))

        with raises(Invalid) as exc:
            validate(data, schema)

        error = exc.value.args[0]
        assert  "-> d -> d did not match 'a'" in error

    def test_multi_pair_non_nested_first(self):
        data = {'a': 'a', 'b':'b', 'c':'c', 'd':'d'}
        schema = (('a', 'b'), ('b', 'b'), ('c', 'c'), ('d', 'd'))

        with raises(Invalid) as exc:
            validate(data, schema)

        error = exc.value.args[0]
        assert  "-> a -> a did not match 'b'" in error

    def test_multi_pair_non_nested_second(self):
        data = {'a': 'a', 'b':'b', 'c':'c', 'd':'d'}
        schema = (('a', 'a'), ('b', 'a'), ('c', 'c'), ('d', 'd'))

        with raises(Invalid) as exc:
            validate(data, schema)

        error = exc.value.args[0]
        assert  "-> b -> b did not match 'a'" in error

    def test_multi_pair_non_nested_third(self):
        data = {'a': 'a', 'b':'b', 'c':'c', 'd':'d'}
        schema = (('a', 'a'), ('b', 'b'), ('c', 'a'), ('d', 'd'))

        with raises(Invalid) as exc:
            validate(data, schema)

        error = exc.value.args[0]
        assert  "-> c -> c did not match 'a'" in error

    def test_multi_pair_non_nested_last_key(self):
        data = {'a': 'a', 'b':'b', 'c':'c', 'd':'d'}
        schema = (('a', 'a'), ('b', 'b'), ('c', 'c'), ('z', 'a'))

        with raises(Invalid) as exc:
            validate(data, schema)

        error = exc.value.args[0]
        assert  "-> d key did not match 'z'" in error

    def test_multi_pair_non_nested_first_key(self):
        data = {'a': 'a', 'b':'b', 'c':'c', 'd':'d'}
        schema = (('aa', 'a'), ('b', 'b'), ('c', 'c'), ('d', 'd'))

        with raises(Invalid) as exc:
            validate(data, schema)

        error = exc.value.args[0]
        assert  "-> a key did not match 'aa'" in error

    def test_multi_pair_non_nested_second_key(self):
        data = {'a': 'a', 'b':'b', 'c':'c', 'd':'d'}
        schema = (('a', 'a'), ('bb', 'b'), ('c', 'c'), ('d', 'd'))

        with raises(Invalid) as exc:
            validate(data, schema)

        error = exc.value.args[0]
        assert  "-> b key did not match 'bb'" in error

    def test_multi_pair_non_nested_third_key(self):
        data = {'a': 'a', 'b':'b', 'c':'c', 'd':'d'}
        schema = (('a', 'a'), ('b', 'b'), ('cc', 'c'), ('d', 'a'))

        with raises(Invalid) as exc:
            validate(data, schema)

        error = exc.value.args[0]
        assert  "-> c key did not match 'cc'" in error

    def test_key_fails_before_value(self):
        data = {'a': {'a': {'b': 'b'}}}
        schema = ('a', ('b', ('b', 'b')))

        with raises(Invalid) as exc:
            validate(data, schema)

        error = exc.value.args[0]
        assert  "-> a -> a key did not match 'b'" in error

class TestDefinedKeys(object):

    # a.k.a. cherry picked validator

    def test_top_level_cherry_pick_with_optionals_passes(self):
        data = {'b': 'b', 'c': 'c', 'd': 'd'}
        schema = (
            ('b', types.string),
            (optional('optional'), types.string)
        )

        assert validate(data, schema, defined_keys=True) is None

    def test_top_level_cherry_pick_with_optionals_complains(self):
        data = {'b': 'b', 'optional': 1, 'garbage': 'yes'}
        schema = (
            ('b', types.string),
            (optional('optional'), types.string)
        )

        with raises(Invalid) as exc:
            validate(data, schema, defined_keys=True)
        error = exc.value.args[0]
        assert '1 did not pass validation against callable: string' in error

    def test_validate_only_one_item(self):
        data = {'b': 'b', 'c': 'c', 'd': 'd'}
        schema = ('d', 'a')

        with raises(Invalid) as exc:
            validate(data, schema, defined_keys=True)

        error = exc.value.args[0]
        assert  "-> d -> d did not match 'a'" in error

    def test_optional_is_kept_optional(self):
        data = {'b': 'b', 'c': 'c', 'd': 'd'}
        schema = (
            ('d', 'd'),
            (optional('f'), 'f'),
        )

        result = validate(data, schema, defined_keys=True)
        assert result is None

    def test_everything_optional_is_allowed(self):
        data = {'b': 'b', 'c': 'c', 'd': 'd'}
        schema = (
            (optional('z'), 'd'),
            (optional('f'), 'f'),
        )

        result = validate(data, schema, defined_keys=True)
        assert result is None

    def test_optional_does_not_affect_cherry_picked(self):
        data = {'b': 'b', 'c': 'c', 'd': 'd'}
        schema = (
            ('d', 'a'),
            (optional('f'), 'f'),
        )

        with raises(Invalid) as exc:
            validate(data, schema, defined_keys=True)

        error = exc.value.args[0]
        assert  "-> d -> d did not match 'a'" in error

class TestCherryPick(object):

    def test_top_level_cherry_pick_with_optionals_passes(self):
        data = {'b': 'b', 'c': 'c', 'd': 'd'}
        schema = cherry_pick((
            ('b', types.string),
            (optional('optional'), types.string))
        )

        assert validate(data, schema) is None

    def test_top_level_cherry_pick_with_optionals_complains(self):
        data = {'b': 'b', 'optional': 1}
        schema = (cherry_pick(
            ('b', types.string),),
            (optional('optional'), types.string)
        )

        with raises(Invalid) as exc:
            validate(data, schema)
        error = exc.value.args[0]
        assert '1 did not pass validation against callable: string' in error

    def test_validate_only_one_item(self):
        data = {'a': {'b': 'b', 'c': 'c', 'd': 'd'}}
        schema = ('a', cherry_pick(('b', 'a')))

        with raises(Invalid) as exc:
            validate(data, schema)

        error = exc.value.args[0]
        assert  "-> a -> b -> b did not match 'a'" in error

    def test_validate_custom_items(self):
        data = {'a': {'b': 'b', 'c': 'c', 'd': 'd'}}
        picky = cherry_pick((('b' ,'b')))
        picky.must_validate = ('b',)
        schema = ('a', picky)

        # should not raise invalid, because 'c' never got validated
        assert validate(data, schema) is None

    def test_complain_about_empty_must_validate(self):
        data = {'a': {'b': 'b', 'c': 'c', 'd': 'd'}}
        picky = cherry_pick((('b' ,'b'), ('c', 'd')))
        picky.must_validate = ()
        schema = ('a', picky)

        # should not raise invalid, because 'c' never got validated
        with raises(SchemaError) as exc:
            validate(data, schema)

        error = exc.value.args[0]
        assert  "-> a  must_validate attribute must not be empty" in error

    def test_complain_about_non_alpha_schema(self):
        data = {'a': 'a', 'b': 'b', 'c': 'c'}
        schema = (('b' ,'b'), ('a', 'a'), ('c', 'c'))

        with raises(SchemaError) as exc:
            validate(data, schema)

        error = exc.value.args[0]
        assert  "b  schema item is not alphabetically ordered" in error

    def test_complain_about_non_alpha_optional_schema(self):
        data = {'a': 'a', 'b': 'b', 'c': 'c'}
        schema = ((optional('b') ,'b'), ('a', 'a'), ('c', 'c'))

        with raises(SchemaError) as exc:
            validate(data, schema)

        error = exc.value.args[0]
        assert  "b  schema item is not alphabetically ordered" in error

class TestHybrid(object):

    def test_hybrid_delegates_when_dict_or_list(self):
        # send off to Recursive when isinstance(value, (dict, list))
        data = {'a':  ['a']}
        hybrid = Hybrid(types.string, iterables.AllItems('b'))
        schema = ('a', hybrid)
        with raises(Invalid) as exc:
            validate(data, schema)
        error = exc.value.args[0]
        assert "list[0] item did not match 'b'" in error

    def test_hybrid_delegates_when_dict_or_empty_list(self):
        # send off to Recursive when isinstance(value, (dict, list))
        data = {'a':  []}
        hybrid = Hybrid(types.string, iterables.AllItems('b'))
        schema = ('a', hybrid)
        assert validate(data, schema) is None


class TestWithIterableValidators(object):

    def test_all_items_pass(self):
        data = {'a': [1, 2, 3, 4, 5]}
        schema = ('a', iterables.AllItems(types.integer))
        assert validate(data, schema) is None

    def test_nested_can_raise_nested_invalid(self):
        data = {'a': {'a': 'b'}}
        data = {'a': [{'a': [1,2]}, {'a': {}}]}
        nested_schema = ('a', iterables.AllItems(types.integer))
        schema = ('a', iterables.AllItems(nested_schema))
        with raises(Invalid) as exc:
            validate(data, schema)
        error = exc.value.args[0]
        assert 'did not pass validation against callable: integer'
        assert exc.value.reason == 'expected a list but got dict'

    def test_nested_raises_nested_invalid(self):
        data = {'a': [{'a': ['a']}, {'b': 'c'}]}
        nested_schema = iterables.MultiIterable(('a', ['b']), ('b', 'c'))
        schema = ('a', iterables.AllItems(nested_schema))
        with raises(Invalid) as exc:
            validate(data, schema)
        assert exc.value.reason == 'expected a list but got dict'

    def test_any_items_pass(self):
        data = {'a': [1, 2, 'a string', 4, 5]}
        schema = ('a', iterables.AnyItem(types.string))
        assert validate(data, schema) is None

    def test_all_items_fail(self):
        data = {'a': [1, 2, '3', 4, 5]}
        schema = ('a', iterables.AllItems(types.integer))
        with raises(Invalid) as exc:
            validate(data, schema)
        error = exc.value.args[0]
        assert 'not of type int' in error
        assert '-> a -> list[2] item did not pass validation against callable: integer' in error

    def test_all_items_fail_length(self):
        data = {'a': [{'a': 2}, {'b': {'a': 'b'}}]}
        schema = ('a', iterables.AllItems((types.string, 2)))
        with raises(SchemaError) as exc:
            validate(data, schema)
        error = exc.value.args[0]
        assert  '-> a -> b  has less items in schema than in data' in error

    def test_all_items_fail_non_callable(self):
        data = {'a': [1, 2, '3', 4, 5]}
        schema = ('a', iterables.AllItems('foo'))
        with raises(Invalid) as exc:
            validate(data, schema)

        error = exc.value.args[0]
        assert  "-> a -> list[0] item did not match 'foo'" in error

    def test_any_items_fail(self):
        data = {'a': [1, 2, 3, 4, 5]}
        schema = ('a', iterables.AnyItem(types.string))
        with raises(Invalid) as exc:
            validate(data, schema)

        error = exc.value.args[0]
        assert  '-> a -> list[] did not contain any valid items against callable: string' in error

    def test_any_items_fail_non_callable(self):
        data = {'a': [1, 2, 3, 4, 5]}
        schema = ('a', iterables.AnyItem('foo'))
        with raises(Invalid) as exc:
            validate(data, schema)

        error = exc.value.args[0]
        assert  "-> a -> list[] did not contain any valid items matching 'foo'" in error

    def test_any_item_with_dictionaries(self):
        data = {'a': [{'a': 1}, {'b': 2}]}
        schema = ('a', iterables.AnyItem(('c', 4)))
        with raises(Invalid) as exc:
            validate(data, schema)

        error = exc.value.args[0]
        assert  "-> a -> list[] did not contain any valid items matching ('c', 4)" in error


class TestWithRecursiveValidators(object):

    def test_all_objects_pass(self):
        data = {'a': {'a': 1, 'b': 2, 'c': 3}}
        schema = ('a', recursive.AllObjects((types.string, types.integer)))
        assert validate(data, schema) is None

    def test_nested_raises_nested_invalid(self):
        data = {'a': {'a': ['a'], 'b': {}}}
        nested_schema = iterables.AllItems(types.string)
        schema = ('a', recursive.AllObjects((types.string, nested_schema)))
        with raises(Invalid) as exc:
            validate(data, schema)
        error = exc.value.args[0]
        assert 'did not pass validation against callable: AllItems' in error

    def test_all_objects_pass_with_hybrid(self):
        data = {'a': {'a': 1, 'b': 2, 'c': 3}}
        dawg = Hybrid(lambda x: True, (types.string, types.integer))
        yo_dawg = recursive.AllObjects(dawg)
        schema = ('a', yo_dawg)
        assert validate(data, schema) is None


    def test_any_objects_pass(self):
        data = {'a': {'a': 1, 'b':'a string', 'c': 3}}
        schema = ('a', recursive.AnyObject((types.string, types.string)))
        assert validate(data, schema) is None

    def test_all_objects_fail(self):
        data = {'a': {'a': 1, 'b': 'a string', 'c': 3}}
        schema = ('a', recursive.AllObjects((types.string, types.integer)))
        with raises(Invalid) as exc:
            validate(data, schema)
        error = exc.value.args[0]
        assert '-> a -> b -> a string did not pass validation against callable: integer' in error
        assert 'not of type int' in error

    def test_all_objects_fail_non_callable(self):
        data = {'a': {'a': 1, 'b': 1, 'c': 1}}
        schema = ('a', recursive.AllObjects((types.string, 2)))
        with raises(Invalid) as exc:
            validate(data, schema)
        error = exc.value.args[0]
        assert  '-> a -> a -> 1 did not match 2' in error

    def test_all_objects_fail_length(self):
        data = {'a': {'a': 2, 'b': {'a': 'b'}}}
        schema = ('a', recursive.AllObjects((types.string, 2)))
        with raises(SchemaError) as exc:
            validate(data, schema)

        error = exc.value.args[0]
        assert  '-> a -> b  has less items in schema than in data' in error

    def test_any_object_fail(self):
        data = {'a': {'a': 1, 'b': 4, 'c': 3}}
        schema = ('a', recursive.AnyObject((types.string, types.string)))
        with raises(Invalid) as exc:
            validate(data, schema)

        error = exc.value.args[0]
        assert  '-> a did not contain any valid objects against callable: AnyObject' in error

    def test_any_objects_fail_non_callable(self):
        data = {'a': {'a': 1, 'b': 4, 'c': 3}}
        schema = ('a', recursive.AnyObject(('a', 'a')))
        with raises(Invalid) as exc:
            validate(data, schema)

        error = exc.value.args[0]
        assert  '-> a did not contain any valid objects against callable: AnyObject' in error

    def test_no_pollution_from_previous_traversing_all_objects(self):
        data = {
            'a' : { 'a': 1, 'b': 2 },
            'b' : { 'c' :1, 'd': 2 },
            'c' : { 'e': 1, 'f': 2 }
        }
        schema = (
            ('a', (('a', 1), ('b', 2))),
            ('b', (('c', 1), ('d', 2))),
            ('c', recursive.AllObjects((types.string, types.string)))
        )
        with raises(Invalid) as exc:
            validate(data, schema)
        error = exc.value.args[0]
        assert '-> c -> e -> 1 did not pass validation against callable: string' in error
        assert 'not of type str' in error

    def test_no_pollution_from_previous_traversing_all_items(self):
        data = {
            'a' : { 'a': 1, 'b': 2 },
            'b' : { 'c' :1, 'd': 2 },
            'c' : ['e', 1, 'f', 2 ]
        }
        schema = (
            ('a', (('a', 1), ('b', 2))),
            ('b', (('c', 1), ('d', 2))),
            ('c', iterables.AllItems(types.string))
        )
        with raises(Invalid) as exc:
            validate(data, schema)
        error = exc.value.args[0]
        assert 'not of type string' in error
        assert '-> c -> list[1] item did not pass validation against callable: string' in error


class TestChainableAllIn(object):

    def test_all_items_pass(self):
        starts = lambda value: True
        data = {'a': 'some string'}
        schema = ('a', chainable.AllIn(types.string, starts))
        assert validate(data, schema) is None

    def test_one_item_fails(self):
        data = {'a': 'some string'}
        schema = ('a', chainable.AllIn(types.string, types.boolean))
        with raises(Invalid) as exc:
            validate(data, schema)
        error = exc.value.args[0]
        assert '-> a -> some string did not pass validation against callable: AllIn -> boolean' in error


class TestChainableAnyIn(object):

    def test_all_items_fail(self):
        def fail(value): raise AssertionError
        data = {'a': 'some string'}
        schema = ('a', chainable.AnyIn(types.boolean, fail))
        with raises(Invalid) as exc:
            validate(data, schema)
        error = exc.value.args[0]
        assert '-> a -> some string did not pass validation against callable: AnyIn' in error

    def test_one_item_passes(self):
        def fail(value): raise AssertionError
        data = {'a': 'some string'}
        schema = ('a', chainable.AnyIn(types.string, fail))
        assert validate(data, schema) is None


class TestOptional(object):

    def test_optional_value(self):
        data = {'optional': '', 'required': 2}
        schema = (('optional', optional(1)), ('required', 2))
        validate(data, schema)

    def test_optional_key(self):
        data = {'required': 2, 'zoo': 1}
        schema = ((optional('optional'), 1), ('required', 2), ('zoo', 1))
        validate(data, schema)



from functools import wraps
from notario.utils import is_callable, safe_repr, ensure


class instance_of(object):
    """
    When trying to make sure the value is coming from any number of valid
    objects, you will want to use this decorator as it will make sure that
    before executing the validator it will comply being of any of the
    ``valid_types``.

    For example, if the input for a given validator can be either a dictionary
    or a list, this validator could be used like::

        from notario import ensure
        @instance_of((list, dict))
        def my_validator(value):
            ensure(len(value) > 0)

    This decorator **needs** to be called as it has a default for valid types,
    which is: ``(list, dict, str)``. A working implementation would look like
    this with the default types::

        from notario import ensure
        @instance_of()
        def my_validator(value):
            ensure(len(value) > 0)

    When it fails, as almost all of Notario's exceptions, it will return
    a meaningful error, this is how passing a boolean to a validator that
    accepts the defaults would raise said error:

    .. doctest:: instance_of

        >>> from notario.decorators import instance_of
        >>> from notario import ensure
        >>> @instance_of()
        ... def my_validator(value):
        ...     ensure(len(value) == 2)
        ...
        >>> my_validator(True)
        Traceback (most recent call last):
        ...
        AssertionError: not of any valid types: ['list', 'dict', 'str']
    """

    def __init__(self, valid_types=None):
        self.valid_types = valid_types or (list, dict, str)
        self.__validator_leaf__ = False
        for t in self.valid_types:
            if t is dict or isinstance(t, dict):
                self.__validator_leaf__ = True

    def __call__(self, func):
        fail_msg = "not of any valid types: %s" % self.valid_names()

        def decorated(value, *a):
            if self.__validator_leaf__ is True:
                if isinstance(value, dict):
                    # convert the 'data' to an actual value. 'data' is of {0: {'key': 'value'}}
                    value = dict(i for i in value.values())
            ensure(isinstance(value, self.valid_types), fail_msg)
            func(value)
        decorated.__validator_leaf__ = self.__validator_leaf__
        return decorated

    def valid_names(self):
        return [safe_repr(obj)
                for obj in self.valid_types]


def not_empty(_object):
    """
    Validates the given input (has to be a valid data structure) is empty.
    Input *has* to be one of: `list`, `dict`, or `string`.

    It is specially useful when most of the validators being created are
    dealing with data structures that should not be empty.
    """
    if is_callable(_object):
        _validator = _object

        @wraps(_validator)
        @instance_of()
        def decorated(value):
            ensure(value, "%s is empty" % safe_repr(value))
            return _validator(value)
        return decorated
    try:
        ensure(len(_object), "%s is empty" % safe_repr(_object))
    except TypeError:
        raise AssertionError("not of any valid types: [list, dict, str]")


def optional(_object):
    """
    This decorator has a double functionality, it can wrap validators and make
    them optional or it can wrap keys and make that entry optional.

    **Optional Validator:**
    Allows to have validators work only when there is a value that contains
    some data, otherwise it will just not pass the information to the actual
    validator and will not fail as a result.

    As any normal decorator, it can be used corectly with the decorator
    syntax or in the actual schema.

    This is how it would look in a schema::

        ('key', optional(my_validator))

    Where ``my_validator`` can be any validator that accepts a single
    argument.

    In case a class based validator is being used (like the ``recursive`` or
    ``iterables`` then it would look like::

        ('key', optional(class_validator(('key', 'value'))))

    Of course, the schema should vary depending on your needs, it is just the
    way of constructing the validator call that should be important.

    **Optional Keys:**
    Sometimes a given data structure may present optional entries. For example
    this data::

        data = {'required': 1, 'optional': 2}

    To represent this, you will need to declare the `optional` key in the
    schema but by wrapping the key with this decorator you will basically tell
    the validation engine that if that key is present it should be validated,
    otherwise, it should be skipped. This is how the schema would look::

        schema = (('required', 1), (optional('optional'), 1))

    The above schema would allow data that is missing the ``optional`` key. The
    data below would pass validation without any issues::

        data = {'required': 1}
    """
    if is_callable(_object):
        validator = _object

        @wraps(validator)
        def decorated(value):
            if value:
                return validator(value)
            return

        return decorated
    else:
        def optional(*args):
            return _object
        optional.is_optional = True
        optional._object = _object
        return optional


def delay(func):
    """
    When schemas are referencing to each other, this decorator will help by
    marking a schema as ``delayed`` to avoid the need for calling a schema to
    generate itself until it is actually needed.

    For example, if a schema function references to itself in this manner::

        def my_schema():
            return (
                ('a', 'foo'),
                ('b', my_schema()),
            )

    Because ``my_schema`` is being called within itself, it will get into
    a recursion problem as soon as it is executed.

    To avoid this, applying the decorator will make it so that the engine will
    acknowledge this is the case, and will *expand* the schema only when it is
    needed. No recursion problems will happen then since we are effectively
    delaying its execution.
    """
    func.__delayed__ = True
    return func



"""
Thread-safe storage for Notario.
"""
from threading import local

_state = local()


def _proxy(key):
    class ObjectProxy(object):
        def __getattr__(self, attr):
            obj = getattr(_state, key)
            return getattr(obj, attr)

        def __setattr__(self, attr, value):
            obj = getattr(_state, key)
            return setattr(obj, attr, value)

        def __delattr__(self, attr):
            obj = getattr(_state, key)
            return delattr(obj, attr)
    return ObjectProxy()


def create_store():
    """
    A helper for setting the _proxy and slapping the store
    object for us.

    :return: A thread-local storage as a dictionary
    """
    new_storage = _proxy('store')
    _state.store = type('store', (object,), {})
    new_storage.store = dict()
    return new_storage.store

store = create_store()


#
# Helpers
#

class new_store(object):

    def __init__(self):
        self.store = create_store()

    def __enter__(self):
        return self.store

    def __exit__( self, exc_type, exc_val, exc_tb):
        # clear the store regardless of exceptions
        del _state.store
        if exc_type:
            return False
        return True


from py.test import raises
from notario import regex


class TestLinker(object):

    def test_do_not_accept_non_pairs(self):
        with raises(TypeError):
            regex.Linker([('a',)], True)

    def test_build_itself_with_same_amount_of_regexes(self):
        rx = regex.Linker([('a', 'some reason'), ('b', 'other reason')], True)
        assert len(rx) == 2

    def test_assertion_error_for_failure(self):
        rx = regex.Linker([(r"\d+", "not a digit")])
        with raises(AssertionError):
            rx('f')

    def test_assertion_error_has_message(self):
        rx = regex.Linker([(r"\w+$", "does not begin with a word")])
        with raises(AssertionError) as exc:
            rx('_*')
        error = exc.value.args[0]
        assert error == 'does not begin with a word'


class TestLinkerSequentially(object):

    def setup(self):
        self.rxs = [(r"^\w+", "begin with a word"),
                    (r"-", "follow a dash"),
                    (r"_$", "end with an underscore")]

    def test_fail_on_first_regex(self):
        rx = regex.Linker(self.rxs)
        with raises(AssertionError) as exc:
            rx(' ')
        error = exc.value.args[0]
        assert error == 'begin with a word'

    def test_fail_on_second_regex(self):
        rx = regex.Linker(self.rxs)
        with raises(AssertionError) as exc:
            rx('f ')
        error = exc.value.args[0]
        assert error == 'follow a dash'

    def test_fail_on_last_regex(self):
        rx = regex.Linker(self.rxs)
        with raises(AssertionError) as exc:
            rx('d- ')
        error = exc.value.args[0]
        assert error == 'end with an underscore'

    def test_fail_on_first_regex_with_prepending(self):
        rx = regex.Linker(self.rxs, True)
        with raises(AssertionError) as exc:
            rx(' ')
        error = exc.value.args[0]
        assert error == 'does not begin with a word'


class TestChain(object):

    def test_prepend_with_multi_regex(self):
        rx = regex.chain((r'^\w+', 'start with a word'),
                         (r'\s+$', 'end with whitespace'))
        with raises(AssertionError) as exc:
            rx('w')
        error = exc.value.args[0]
        assert error == 'does not end with whitespace'

    def test_dont_prepend_with_multi_regex(self):
        rx = regex.chain((r'^\w+', 'start with a word'),
                         (r'\s+$', 'end with whitespace'),
                         prepend_negation=False)
        with raises(AssertionError) as exc:
            rx('w')
        error = exc.value.args[0]
        assert error == 'end with whitespace'




"""
Chainable validators are *encapsulating* validators. They
usually will not validate per se, but can contain other validators
inside them and pass the value to them.
"""
from notario.utils import is_callable


class BasicChainValidator(object):
    """
    The base chainable validator, should not be used directly
    but can be sub-classed to extend into custom chainable
    validators.
    """

    def __init__(self, *args):
        for i in args:
            if not is_callable(i):
                raise TypeError("got a non-callable argument: %s" % repr(i))
        self.args = args


class AllIn(BasicChainValidator):
    """
    Validates against all the validators passed in. This chainable validator
    will pass in the actual to every single validator that is contained as an
    argument.

    Example usage::

        from notario.validators import types

        data = {'foo' : "some string"}
        schema = ('foo', AllIn(types.string))
        validate(data, schema)

    When more than one validator needs to be chained this validator can take it
    in as another argument. Lets say that you have a validator that specifies
    a minimum length a maxium length and that it starts with the letter 's'::::

        data = {'foo' : "some string"}
        schema = ('foo', AllIn(min_length, max_length, StartsWith('s'))
        validate(data, schema)


    :raises: TypeError if the validator is *not* a callable
    """

    def __call__(self, value):
        for validator in self.args:
            try:
                validator(value)
            except AssertionError as exc:
                self.__name__ = 'AllIn -> %s' % validator.__name__
                raise AssertionError(exc)


class AnyIn(BasicChainValidator):
    """
    If any contained validator passes it skips any others, even if those others
    might fail at some point.  If no validators pass at the end it fails
    pointing out that the ``AnyIn`` validator was not able to pass against any
    contained validator.

    :raises: TypeError if the validator is *not* a callable
    """
    __name__ = 'AnyIn'

    def __call__(self, value):
        for validator in self.args:
            try:
                return validator(value)
            except AssertionError:
                pass

        raise AssertionError("did not passed validation against any validator")



from notario import normal


class TestSchema(object):

    def test_return_data_structure_with_cero_index(self):
        result = normal.Schema({}, ['a']).normalized()
        assert result == {0: ['a']}

    def test_recurse_if_second_item_is_typle(self):
        data = ('a', ('a', 'b'))
        result = normal.Schema({}, data).normalized()
        assert result == {0: ('a', {0: ('a', 'b')})}

    def test_respect_more_than_two_values_in_tuple(self):
        data = ('a', (('a', 'b'), ('c', 'c'), ('d', 'd')))
        result = normal.Schema({}, data).normalized()
        assert result == {0: ('a', {0: ('a', 'b'), 1: ('c', 'c'), 2: ('d', 'd')})}

    def test_basic_key_value_pairs(self):
        data = (('a', 'b'), ('b', 'b'), ('c', 'c'))
        result = normal.Schema({}, data).normalized()
        assert result == {0: ('a', 'b'), 1: ('b', 'b'), 2: ('c', 'c')}



from pytest import raises
from notario.validators import chainable
from notario.validators import types


class TestBasicChainValidator(object):

    def test_complains_when_args_are_not_callable(self):
        with raises(TypeError) as exc:
            chainable.BasicChainValidator("not a callable")
        assert exc.value.args[0] == "got a non-callable argument: 'not a callable'"

    def test_assigns_validators_if_all_are_callable(self):
        def foo(): pass
        def bar(): pass
        chain = chainable.BasicChainValidator(foo, bar)
        assert chain.args == (foo, bar)


class TestAllIn(object):

    def test_all_validators_pass(self):
        def foo(value): pass
        def bar(value): pass
        chain = chainable.AllIn(foo, bar)
        assert chain('some value') is None

    def test_a_validator_fails_with_a_tree_path(self):
        chain = chainable.AllIn(types.boolean, types.string)
        with raises(AssertionError):
            chain("some string")


class TestAnyIn(object):

    def test_all_validators_pass(self):
        def foo(value): pass
        def bar(value): pass
        chain = chainable.AnyIn(foo, bar)
        assert chain('some value') is None

    def test_no_validator_passes(self):
        def foo(value): raise AssertionError
        def bar(value): raise AssertionError
        chain = chainable.AnyIn(foo, bar)
        with raises(AssertionError) as exc:
            chain('some value')
        assert exc.value.args[0] == 'did not passed validation against any validator'
